#
#    Copyright 2015 IBM Corp.
#
#    All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation


class PoolMember(neutron.NeutronResource):
    """A resource for managing LBaaS v2 Pool Members.

    A pool member represents a single backend node.
    """

    support_status = support.SupportStatus(version='6.0.0')

    required_service_extension = 'lbaasv2'

    entity = 'lbaas_member'

    res_info_key = 'member'

    PROPERTIES = (
        POOL, ADDRESS, PROTOCOL_PORT, WEIGHT, ADMIN_STATE_UP,
        SUBNET,
    ) = (
        'pool', 'address', 'protocol_port', 'weight', 'admin_state_up',
        'subnet'
    )

    ATTRIBUTES = (
        ADDRESS_ATTR, POOL_ID_ATTR
    ) = (
        'address', 'pool_id'
    )

    properties_schema = {
        POOL: properties.Schema(
            properties.Schema.STRING,
            _('Name or ID of the load balancing pool.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('neutron.lbaas.pool')
            ]
        ),
        ADDRESS: properties.Schema(
            properties.Schema.STRING,
            _('IP address of the pool member on the pool network.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('ip_addr')
            ]
        ),
        PROTOCOL_PORT: properties.Schema(
            properties.Schema.INTEGER,
            _('Port on which the pool member listens for requests or '
              'connections.'),
            required=True,
            constraints=[
                constraints.Range(1, 65535),
            ]
        ),
        WEIGHT: properties.Schema(
            properties.Schema.INTEGER,
            _('Weight of pool member in the pool (default to 1).'),
            default=1,
            constraints=[
                constraints.Range(0, 256),
            ],
            update_allowed=True
        ),
        ADMIN_STATE_UP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('The administrative state of the pool member.'),
            default=True,
            update_allowed=True
        ),
        SUBNET: properties.Schema(
            properties.Schema.STRING,
            _('Subnet name or ID of this member.'),
            constraints=[
                constraints.CustomConstraint('neutron.subnet')
            ],
            # Make this required untill bug #1585100 is resolved.
            required=True
        ),
    }

    attributes_schema = {
        ADDRESS_ATTR: attributes.Schema(
            _('The IP address of the pool member.'),
            type=attributes.Schema.STRING
        ),
        POOL_ID_ATTR: attributes.Schema(
            _('The ID of the pool to which the pool member belongs.'),
            type=attributes.Schema.STRING
        )
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.SUBNET],
                client_plugin=self.client_plugin(),
                finder='find_resourceid_by_name_or_id',
                entity='subnet'
            ),
        ]

    def __init__(self, name, definition, stack):
        super(PoolMember, self).__init__(name, definition, stack)
        self._pool_id = None
        self._lb_id = None

    @property
    def pool_id(self):
        if self._pool_id is None:
            client_plugin = self.client_plugin()
            self._pool_id = client_plugin.find_resourceid_by_name_or_id(
                client_plugin.RES_TYPE_LB_POOL,
                self.properties[self.POOL])
        return self._pool_id

    @property
    def lb_id(self):
        if self._lb_id is None:
            pool = self.client().show_lbaas_pool(self.pool_id)['pool']

            listener_id = pool['listeners'][0]['id']
            listener = self.client().show_listener(listener_id)['listener']

            self._lb_id = listener['loadbalancers'][0]['id']
        return self._lb_id

    def _check_lb_status(self):
        return self.client_plugin().check_lb_status(self.lb_id)

    def handle_create(self):
        properties = self.prepare_properties(
            self.properties,
            self.physical_resource_name())

        self.client_plugin().resolve_pool(
            properties, self.POOL, 'pool_id')
        properties.pop('pool_id')
        properties['subnet_id'] = properties.pop(self.SUBNET)
        return properties

    def check_create_complete(self, properties):
        if self.resource_id is None:
            try:
                member = self.client().create_lbaas_member(
                    self.pool_id, {'member': properties})['member']
                self.resource_id_set(member['id'])
            except Exception as ex:
                if self.client_plugin().is_invalid(ex):
                    return False
                raise

        return self._check_lb_status()

    def _res_get_args(self):
        return [self.resource_id, self.pool_id]

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self._update_called = False
        return prop_diff

    def check_update_complete(self, prop_diff):
        if not prop_diff:
            return True

        if not self._update_called:
            try:
                self.client().update_lbaas_member(self.resource_id,
                                                  self.pool_id,
                                                  {'member': prop_diff})
                self._update_called = True
            except Exception as ex:
                if self.client_plugin().is_invalid(ex):
                    return False
                raise

        return self._check_lb_status()

    def handle_delete(self):
        self._delete_called = False

    def check_delete_complete(self, data):
        if self.resource_id is None:
            return True

        if not self._delete_called:
            try:
                self.client().delete_lbaas_member(self.resource_id,
                                                  self.pool_id)
                self._delete_called = True
            except Exception as ex:
                if self.client_plugin().is_invalid(ex):
                    return False
                elif self.client_plugin().is_not_found(ex):
                    return True
                raise

        return self._check_lb_status()


def resource_mapping():
    return {
        'OS::Neutron::LBaaS::PoolMember': PoolMember,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from requests import exceptions

from heat.common import exception
from heat.common.i18n import _
from heat.common import template_format
from heat.common import urlfetch
from heat.engine import attributes
from heat.engine import properties
from heat.engine.resources import stack_resource


class NestedStack(stack_resource.StackResource):
    """Represents a child stack to allow composition of templates."""

    PROPERTIES = (
        TEMPLATE_URL, TIMEOUT_IN_MINS, PARAMETERS,
    ) = (
        'TemplateURL', 'TimeoutInMinutes', 'Parameters',
    )

    properties_schema = {
        TEMPLATE_URL: properties.Schema(
            properties.Schema.STRING,
            _('The URL of a template that specifies the stack to be created '
              'as a resource.'),
            required=True,
            update_allowed=True
        ),
        TIMEOUT_IN_MINS: properties.Schema(
            properties.Schema.INTEGER,
            _('The length of time, in minutes, to wait for the nested stack '
              'creation.'),
            update_allowed=True
        ),
        PARAMETERS: properties.Schema(
            properties.Schema.MAP,
            _('The set of parameters passed to this nested stack.'),
            update_allowed=True
        ),
    }

    def child_template(self):
        template_url = self.properties[self.TEMPLATE_URL]
        try:
            template_data = urlfetch.get(template_url)
        except (exceptions.RequestException, IOError) as r_exc:
            raise ValueError(_("Could not fetch remote template '%(url)s': "
                             "%(exc)s") %
                             {'url': template_url,
                              'exc': r_exc})

        return template_format.parse(template_data, template_url)

    def child_params(self):
        return self.properties[self.PARAMETERS]

    def handle_adopt(self, resource_data=None):
        return self._create_with_template(resource_adopt_data=resource_data)

    def handle_create(self):
        return self._create_with_template()

    def _create_with_template(self, resource_adopt_data=None):
        template = self.child_template()
        return self.create_with_template(template,
                                         self.child_params(),
                                         self.properties[self.TIMEOUT_IN_MINS],
                                         adopt_data=resource_adopt_data)

    def get_attribute(self, key, *path):
        if key and not key.startswith('Outputs.'):
            raise exception.InvalidTemplateAttribute(resource=self.name,
                                                     key=key)
        try:
            attribute = self.get_output(key.partition('.')[-1])
        except exception.NotFound:
            raise exception.InvalidTemplateAttribute(resource=self.name,
                                                     key=key)
        return attributes.select_from_attribute(attribute, path)

    def get_reference_id(self):
        identifier = self.nested_identifier()
        if identifier is None:
            return str(self.name)

        return identifier.arn()

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        # Nested stack template may be changed even if the prop_diff is empty.
        self.properties = json_snippet.properties(self.properties_schema,
                                                  self.context)
        return self.update_with_template(self.child_template(),
                                         self.properties[self.PARAMETERS],
                                         self.properties[self.TIMEOUT_IN_MINS])


def resource_mapping():
    return {
        'AWS::CloudFormation::Stack': NestedStack,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import service_utils
from heat.engine import stack_lock
from heat.objects import stack as stack_object
from heat.objects import stack_lock as stack_lock_object
from heat.tests import common
from heat.tests import utils


class StackLockTest(common.HeatTestCase):
    def setUp(self):
        super(StackLockTest, self).setUp()
        self.context = utils.dummy_context()
        self.stack_id = "aae01f2d-52ae-47ac-8a0d-3fde3d220fea"
        self.engine_id = service_utils.generate_engine_id()
        stack = mock.MagicMock()
        stack.id = self.stack_id
        stack.name = "test_stack"
        stack.action = "CREATE"
        self.mock_get_by_id = self.patchobject(
            stack_object.Stack, 'get_by_id', return_value=stack)

    class TestThreadLockException(Exception):
        pass

    def test_successful_acquire_new_lock(self):
        mock_create = self.patchobject(stack_lock_object.StackLock,
                                       'create',
                                       return_value=None)

        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        slock.acquire()

        mock_create.assert_called_once_with(
            self.context, self.stack_id, self.engine_id)

    def test_failed_acquire_existing_lock_current_engine(self):
        mock_create = self.patchobject(stack_lock_object.StackLock,
                                       'create',
                                       return_value=self.engine_id)

        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)

        self.assertRaises(exception.ActionInProgress, slock.acquire)
        self.mock_get_by_id.assert_called_once_with(
            self.context,
            self.stack_id,
            show_deleted=True,
            eager_load=False)
        mock_create.assert_called_once_with(
            self.context, self.stack_id, self.engine_id)

    def test_successful_acquire_existing_lock_engine_dead(self):
        mock_create = self.patchobject(stack_lock_object.StackLock,
                                       'create',
                                       return_value='fake-engine-id')
        mock_steal = self.patchobject(stack_lock_object.StackLock,
                                      'steal',
                                      return_value=None)

        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        self.patchobject(service_utils, 'engine_alive', return_value=False)
        slock.acquire()

        mock_create.assert_called_once_with(
            self.context, self.stack_id, self.engine_id)
        mock_steal.assert_called_once_with(
            self.context, self.stack_id, 'fake-engine-id', self.engine_id)

    def test_failed_acquire_existing_lock_engine_alive(self):
        mock_create = self.patchobject(stack_lock_object.StackLock,
                                       'create',
                                       return_value='fake-engine-id')

        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        self.patchobject(service_utils, 'engine_alive', return_value=True)
        self.assertRaises(exception.ActionInProgress, slock.acquire)
        self.mock_get_by_id.assert_called_once_with(
            self.context,
            self.stack_id,
            show_deleted=True,
            eager_load=False)

        mock_create.assert_called_once_with(
            self.context, self.stack_id, self.engine_id)

    def test_failed_acquire_existing_lock_engine_dead(self):
        mock_create = self.patchobject(stack_lock_object.StackLock,
                                       'create',
                                       return_value='fake-engine-id')
        mock_steal = self.patchobject(stack_lock_object.StackLock,
                                      'steal',
                                      return_value='fake-engine-id2')

        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        self.patchobject(service_utils, 'engine_alive', return_value=False)
        self.assertRaises(exception.ActionInProgress, slock.acquire)
        self.mock_get_by_id.assert_called_once_with(
            self.context,
            self.stack_id,
            show_deleted=True,
            eager_load=False)

        mock_create.assert_called_once_with(
            self.context, self.stack_id, self.engine_id)
        mock_steal.assert_called_once_with(
            self.context, self.stack_id, 'fake-engine-id', self.engine_id)

    def test_successful_acquire_with_retry(self):
        mock_create = self.patchobject(stack_lock_object.StackLock,
                                       'create',
                                       return_value='fake-engine-id')
        mock_steal = self.patchobject(stack_lock_object.StackLock,
                                      'steal',
                                      side_effect=[True, None])

        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        self.patchobject(service_utils, 'engine_alive', return_value=False)
        slock.acquire()

        mock_create.assert_has_calls(
            [mock.call(self.context, self.stack_id, self.engine_id)] * 2)
        mock_steal.assert_has_calls(
            [mock.call(self.context, self.stack_id,
                       'fake-engine-id', self.engine_id)] * 2)

    def test_failed_acquire_one_retry_only(self):
        mock_create = self.patchobject(stack_lock_object.StackLock,
                                       'create',
                                       return_value='fake-engine-id')
        mock_steal = self.patchobject(stack_lock_object.StackLock,
                                      'steal',
                                      return_value=True)

        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        self.patchobject(service_utils, 'engine_alive', return_value=False)
        self.assertRaises(exception.ActionInProgress, slock.acquire)
        self.mock_get_by_id.assert_called_with(
            self.context,
            self.stack_id,
            show_deleted=True,
            eager_load=False)

        mock_create.assert_has_calls(
            [mock.call(self.context, self.stack_id, self.engine_id)] * 2)
        mock_steal.assert_has_calls(
            [mock.call(self.context, self.stack_id,
                       'fake-engine-id', self.engine_id)] * 2)

    def test_context_mgr_exception(self):
        stack_lock_object.StackLock.create = mock.Mock(return_value=None)
        stack_lock_object.StackLock.release = mock.Mock(return_value=None)
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)

        def check_lock():
            with slock:
                self.assertEqual(1,
                                 stack_lock_object.StackLock.create.call_count)
                raise self.TestThreadLockException
        self.assertRaises(self.TestThreadLockException, check_lock)
        self.assertEqual(1, stack_lock_object.StackLock.release.call_count)

    def test_context_mgr_noexception(self):
        stack_lock_object.StackLock.create = mock.Mock(return_value=None)
        stack_lock_object.StackLock.release = mock.Mock(return_value=None)
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)

        with slock:
            self.assertEqual(1,
                             stack_lock_object.StackLock.create.call_count)

        self.assertEqual(1, stack_lock_object.StackLock.release.call_count)

    def test_thread_lock_context_mgr_exception_acquire_success(self):
        stack_lock_object.StackLock.create = mock.Mock(return_value=None)
        stack_lock_object.StackLock.release = mock.Mock(return_value=None)
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)

        def check_thread_lock():
            with slock.thread_lock():
                self.assertEqual(1,
                                 stack_lock_object.StackLock.create.call_count)
                raise self.TestThreadLockException
        self.assertRaises(self.TestThreadLockException, check_thread_lock)
        self.assertEqual(1, stack_lock_object.StackLock.release.call_count)

    def test_thread_lock_context_mgr_exception_acquire_fail(self):
        stack_lock_object.StackLock.create = mock.Mock(
            return_value=self.engine_id)
        stack_lock_object.StackLock.release = mock.Mock()
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)

        def check_thread_lock():
            with slock.thread_lock():
                self.assertEqual(1,
                                 stack_lock_object.StackLock.create.call_count)
                raise exception.ActionInProgress
        self.assertRaises(exception.ActionInProgress, check_thread_lock)
        self.assertFalse(stack_lock_object.StackLock.release.called)

    def test_thread_lock_context_mgr_no_exception(self):
        stack_lock_object.StackLock.create = mock.Mock(return_value=None)
        stack_lock_object.StackLock.release = mock.Mock(return_value=None)
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        with slock.thread_lock():
            self.assertEqual(1, stack_lock_object.StackLock.create.call_count)
        self.assertFalse(stack_lock_object.StackLock.release.called)

    def test_try_thread_lock_context_mgr_exception(self):
        stack_lock_object.StackLock.create = mock.Mock(return_value=None)
        stack_lock_object.StackLock.release = mock.Mock(return_value=None)
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)

        def check_thread_lock():
            with slock.try_thread_lock():
                self.assertEqual(1,
                                 stack_lock_object.StackLock.create.call_count)
                raise self.TestThreadLockException
        self.assertRaises(self.TestThreadLockException, check_thread_lock)
        self.assertEqual(1, stack_lock_object.StackLock.release.call_count)

    def test_try_thread_lock_context_mgr_no_exception(self):
        stack_lock_object.StackLock.create = mock.Mock(return_value=None)
        stack_lock_object.StackLock.release = mock.Mock(return_value=None)
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)
        with slock.try_thread_lock():
            self.assertEqual(1, stack_lock_object.StackLock.create.call_count)
        self.assertFalse(stack_lock_object.StackLock.release.called)

    def test_try_thread_lock_context_mgr_existing_lock(self):
        stack_lock_object.StackLock.create = mock.Mock(return_value=1234)
        stack_lock_object.StackLock.release = mock.Mock(return_value=None)
        slock = stack_lock.StackLock(self.context, self.stack_id,
                                     self.engine_id)

        def check_thread_lock():
            with slock.try_thread_lock():
                self.assertEqual(1,
                                 stack_lock_object.StackLock.create.call_count)
                raise self.TestThreadLockException
        self.assertRaises(self.TestThreadLockException, check_thread_lock)
        self.assertFalse(stack_lock_object.StackLock.release.called)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import netutils

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation


class Subnet(neutron.NeutronResource):
    """A resource for managing Neutron subnets.

    A subnet represents an IP address block that can be used for assigning IP
    addresses to virtual instances. Each subnet must have a CIDR and must be
    associated with a network. IPs can be either selected from the whole subnet
    CIDR, or from "allocation pools" that can be specified by the user.
    """

    entity = 'subnet'

    PROPERTIES = (
        NETWORK_ID, NETWORK, SUBNETPOOL, PREFIXLEN, CIDR,
        VALUE_SPECS, NAME, IP_VERSION, DNS_NAMESERVERS, GATEWAY_IP,
        ENABLE_DHCP, ALLOCATION_POOLS, TENANT_ID, HOST_ROUTES,
        IPV6_RA_MODE, IPV6_ADDRESS_MODE, SEGMENT, TAGS,
    ) = (
        'network_id', 'network', 'subnetpool', 'prefixlen', 'cidr',
        'value_specs', 'name', 'ip_version', 'dns_nameservers', 'gateway_ip',
        'enable_dhcp', 'allocation_pools', 'tenant_id', 'host_routes',
        'ipv6_ra_mode', 'ipv6_address_mode', 'segment', 'tags',
    )

    _ALLOCATION_POOL_KEYS = (
        ALLOCATION_POOL_START, ALLOCATION_POOL_END,
    ) = (
        'start', 'end',
    )

    _HOST_ROUTES_KEYS = (
        ROUTE_DESTINATION, ROUTE_NEXTHOP,
    ) = (
        'destination', 'nexthop',
    )

    _IPV6_DHCP_MODES = (
        DHCPV6_STATEFUL, DHCPV6_STATELESS, SLAAC,
    ) = (
        'dhcpv6-stateful', 'dhcpv6-stateless', 'slaac',
    )

    ATTRIBUTES = (
        NAME_ATTR, NETWORK_ID_ATTR, TENANT_ID_ATTR, ALLOCATION_POOLS_ATTR,
        GATEWAY_IP_ATTR, HOST_ROUTES_ATTR, IP_VERSION_ATTR, CIDR_ATTR,
        DNS_NAMESERVERS_ATTR, ENABLE_DHCP_ATTR,
    ) = (
        'name', 'network_id', 'tenant_id', 'allocation_pools',
        'gateway_ip', 'host_routes', 'ip_version', 'cidr',
        'dns_nameservers', 'enable_dhcp',
    )

    properties_schema = {
        NETWORK_ID: properties.Schema(
            properties.Schema.STRING,
            support_status=support.SupportStatus(
                status=support.HIDDEN,
                version='5.0.0',
                message=_('Use property %s.') % NETWORK,
                previous_status=support.SupportStatus(
                    status=support.DEPRECATED,
                    version='2014.2'
                )
            ),
            constraints=[
                constraints.CustomConstraint('neutron.network')
            ],
        ),
        NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('The ID of the attached network.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('neutron.network')
            ],
            support_status=support.SupportStatus(version='2014.2')
        ),
        SUBNETPOOL: properties.Schema(
            properties.Schema.STRING,
            _('The name or ID of the subnet pool.'),
            constraints=[
                constraints.CustomConstraint('neutron.subnetpool')
            ],
            support_status=support.SupportStatus(version='6.0.0'),
        ),
        PREFIXLEN: properties.Schema(
            properties.Schema.INTEGER,
            _('Prefix length for subnet allocation from subnet pool.'),
            constraints=[constraints.Range(min=0)],
            support_status=support.SupportStatus(version='6.0.0'),
        ),
        CIDR: properties.Schema(
            properties.Schema.STRING,
            _('The CIDR.'),
            constraints=[
                constraints.CustomConstraint('net_cidr')
            ]
        ),
        VALUE_SPECS: properties.Schema(
            properties.Schema.MAP,
            _('Extra parameters to include in the request.'),
            default={},
            update_allowed=True
        ),
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('The name of the subnet.'),
            update_allowed=True
        ),
        IP_VERSION: properties.Schema(
            properties.Schema.INTEGER,
            _('The IP version, which is 4 or 6.'),
            default=4,
            constraints=[
                constraints.AllowedValues([4, 6]),
            ]
        ),
        DNS_NAMESERVERS: properties.Schema(
            properties.Schema.LIST,
            _('A specified set of DNS name servers to be used.'),
            default=[],
            update_allowed=True
        ),
        GATEWAY_IP: properties.Schema(
            properties.Schema.STRING,
            _('The gateway IP address. Set to any of [ null | ~ | "" ] '
              'to create/update a subnet without a gateway. '
              'If omitted when creation, neutron will assign the first '
              'free IP address within the subnet to the gateway '
              'automatically. If remove this from template when update, '
              'the old gateway IP address will be detached.'),
            update_allowed=True
        ),
        ENABLE_DHCP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Set to true if DHCP is enabled and false if DHCP is disabled.'),
            default=True,
            update_allowed=True
        ),
        ALLOCATION_POOLS: properties.Schema(
            properties.Schema.LIST,
            _('The start and end addresses for the allocation pools.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    ALLOCATION_POOL_START: properties.Schema(
                        properties.Schema.STRING,
                        _('Start address for the allocation pool.'),
                        required=True,
                        constraints=[
                            constraints.CustomConstraint('ip_addr')
                        ]
                    ),
                    ALLOCATION_POOL_END: properties.Schema(
                        properties.Schema.STRING,
                        _('End address for the allocation pool.'),
                        required=True,
                        constraints=[
                            constraints.CustomConstraint('ip_addr')
                        ]
                    ),
                },
            ),
            update_allowed=True
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('The ID of the tenant who owns the network. Only administrative '
              'users can specify a tenant ID other than their own.')
        ),
        HOST_ROUTES: properties.Schema(
            properties.Schema.LIST,
            _('A list of host route dictionaries for the subnet.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    ROUTE_DESTINATION: properties.Schema(
                        properties.Schema.STRING,
                        _('The destination for static route.'),
                        required=True,
                        constraints=[
                            constraints.CustomConstraint('net_cidr')
                        ]
                    ),
                    ROUTE_NEXTHOP: properties.Schema(
                        properties.Schema.STRING,
                        _('The next hop for the destination.'),
                        required=True,
                        constraints=[
                            constraints.CustomConstraint('ip_addr')
                        ]
                    ),
                },
            ),
            update_allowed=True
        ),
        IPV6_RA_MODE: properties.Schema(
            properties.Schema.STRING,
            _('IPv6 RA (Router Advertisement) mode.'),
            constraints=[
                constraints.AllowedValues([DHCPV6_STATEFUL, DHCPV6_STATELESS,
                                           SLAAC]),
            ],
            support_status=support.SupportStatus(version='2015.1')
        ),
        IPV6_ADDRESS_MODE: properties.Schema(
            properties.Schema.STRING,
            _('IPv6 address mode.'),
            constraints=[
                constraints.AllowedValues([DHCPV6_STATEFUL, DHCPV6_STATELESS,
                                           SLAAC]),
            ],
            support_status=support.SupportStatus(version='2015.1')
        ),
        SEGMENT: properties.Schema(
            properties.Schema.STRING,
            _('The name/ID of the segment to associate.'),
            constraints=[
                constraints.CustomConstraint('neutron.segment')
            ],
            update_allowed=True,
            support_status=support.SupportStatus(
                version='11.0.0',
                status=support.SUPPORTED,
                message=_('Update allowed since version 11.0.0.'),
                previous_status=support.SupportStatus(
                    version='9.0.0',
                    status=support.SUPPORTED
                )
            )
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('The tags to be added to the subnet.'),
            schema=properties.Schema(properties.Schema.STRING),
            update_allowed=True,
            support_status=support.SupportStatus(version='9.0.0')
        ),
    }

    attributes_schema = {
        NAME_ATTR: attributes.Schema(
            _("Friendly name of the subnet."),
            type=attributes.Schema.STRING
        ),
        NETWORK_ID_ATTR: attributes.Schema(
            _("Parent network of the subnet."),
            type=attributes.Schema.STRING
        ),
        TENANT_ID_ATTR: attributes.Schema(
            _("Tenant owning the subnet."),
            type=attributes.Schema.STRING
        ),
        ALLOCATION_POOLS_ATTR: attributes.Schema(
            _("Ip allocation pools and their ranges."),
            type=attributes.Schema.LIST
        ),
        GATEWAY_IP_ATTR: attributes.Schema(
            _("Ip of the subnet's gateway."),
            type=attributes.Schema.STRING
        ),
        HOST_ROUTES_ATTR: attributes.Schema(
            _("Additional routes for this subnet."),
            type=attributes.Schema.LIST
        ),
        IP_VERSION_ATTR: attributes.Schema(
            _("Ip version for the subnet."),
            type=attributes.Schema.STRING
        ),
        CIDR_ATTR: attributes.Schema(
            _("CIDR block notation for this subnet."),
            type=attributes.Schema.STRING
        ),
        DNS_NAMESERVERS_ATTR: attributes.Schema(
            _("List of dns nameservers."),
            type=attributes.Schema.LIST
        ),
        ENABLE_DHCP_ATTR: attributes.Schema(
            _("'true' if DHCP is enabled for this subnet; 'false' otherwise."),
            type=attributes.Schema.STRING
        ),
    }

    def translation_rules(self, props):
        client_plugin = self.client_plugin()
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.REPLACE,
                [self.NETWORK],
                value_path=[self.NETWORK_ID]),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.NETWORK],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_NETWORK
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.SUBNETPOOL],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_SUBNET_POOL
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.SEGMENT],
                client_plugin=self.client_plugin('openstack'),
                finder='find_network_segment'
            )
        ]

    @classmethod
    def _null_gateway_ip(cls, props):
        if cls.GATEWAY_IP not in props:
            return
        # Specifying null in the gateway_ip will result in
        # a property containing an empty string.
        # A null gateway_ip has special meaning in the API
        # so this needs to be set back to None.
        # See bug https://bugs.launchpad.net/heat/+bug/1226666
        if props.get(cls.GATEWAY_IP) == '':
            props[cls.GATEWAY_IP] = None

    def validate(self):
        super(Subnet, self).validate()
        subnetpool = self.properties[self.SUBNETPOOL]
        prefixlen = self.properties[self.PREFIXLEN]
        cidr = self.properties[self.CIDR]
        if subnetpool is not None and cidr:
            raise exception.ResourcePropertyConflict(self.SUBNETPOOL,
                                                     self.CIDR)
        if subnetpool is None and not cidr:
            raise exception.PropertyUnspecifiedError(self.SUBNETPOOL,
                                                     self.CIDR)
        if prefixlen and cidr:
            raise exception.ResourcePropertyConflict(self.PREFIXLEN,
                                                     self.CIDR)
        ra_mode = self.properties[self.IPV6_RA_MODE]
        address_mode = self.properties[self.IPV6_ADDRESS_MODE]

        if (self.properties[self.IP_VERSION] == 4) and (
                ra_mode or address_mode):
            msg = _('ipv6_ra_mode and ipv6_address_mode are not supported '
                    'for ipv4.')
            raise exception.StackValidationFailed(message=msg)
        if ra_mode and address_mode and (ra_mode != address_mode):
            msg = _('When both ipv6_ra_mode and ipv6_address_mode are set, '
                    'they must be equal.')
            raise exception.StackValidationFailed(message=msg)

        gateway_ip = self.properties.get(self.GATEWAY_IP)
        if (gateway_ip and gateway_ip not in ['~', ''] and
                not netutils.is_valid_ip(gateway_ip)):
            msg = (_('Gateway IP address "%(gateway)s" is in '
                     'invalid format.'), gateway_ip)
            raise exception.StackValidationFailed(message=msg)

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())
        props['network_id'] = props.pop(self.NETWORK)
        if self.SEGMENT in props and props[self.SEGMENT]:
            props['segment_id'] = props.pop(self.SEGMENT)

        tags = props.pop(self.TAGS, [])

        if self.SUBNETPOOL in props and props[self.SUBNETPOOL]:
            props['subnetpool_id'] = props.pop('subnetpool')
        self._null_gateway_ip(props)

        subnet = self.client().create_subnet({'subnet': props})['subnet']
        self.resource_id_set(subnet['id'])

        if tags:
            self.set_tags(tags)

    def handle_delete(self):
        try:
            self.client().delete_subnet(self.resource_id)
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
        else:
            return True

    def _validate_segment_update_supported(self):
        # TODO(hjensas): Validation to ensure the subnet-segmentid-writable
        # extension is available.
        # https://storyboard.openstack.org/#!/story/2002189
        # Current segment id must be None
        if self.properties[self.SEGMENT] is not None:
            msg = _('Updating the subnet segment assciation only allowed '
                    'when the current segment_id is None. The subnet is '
                    'currently associated with segment. In this state update')
            raise exception.ResourceActionNotSupported(action=msg)
        else:
            return True

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.prepare_update_properties(prop_diff)
            if (self.ALLOCATION_POOLS in prop_diff and
                    prop_diff[self.ALLOCATION_POOLS] is None):
                prop_diff[self.ALLOCATION_POOLS] = []
            if (self.SEGMENT in prop_diff and prop_diff[self.SEGMENT] and
                    self._validate_segment_update_supported()):
                prop_diff['segment_id'] = prop_diff.pop(self.SEGMENT)

            # If the new value is '', set to None
            self._null_gateway_ip(prop_diff)
            if self.TAGS in prop_diff:
                tags = prop_diff.pop(self.TAGS)
                self.set_tags(tags)
            self.client().update_subnet(
                self.resource_id, {'subnet': prop_diff})


def resource_mapping():
    return {
        'OS::Neutron::Subnet': Subnet,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources import server_base
from heat.engine import support

cfg.CONF.import_opt('default_software_config_transport', 'heat.common.config')
cfg.CONF.import_opt('default_user_data_format', 'heat.common.config')
cfg.CONF.import_opt('max_server_name_length', 'heat.common.config')

LOG = logging.getLogger(__name__)


class DeployedServer(server_base.BaseServer):
    """A resource for managing servers that are already deployed.

    A DeployedServer resource manages resources for servers that have been
    deployed externally from OpenStack. These servers can be associated with
    SoftwareDeployments for further orchestration via Heat.
    """

    PROPERTIES = (
        NAME, METADATA, SOFTWARE_CONFIG_TRANSPORT,
        DEPLOYMENT_SWIFT_DATA
    ) = (
        'name', 'metadata', 'software_config_transport',
        'deployment_swift_data'
    )

    _SOFTWARE_CONFIG_TRANSPORTS = (
        POLL_SERVER_CFN, POLL_SERVER_HEAT, POLL_TEMP_URL, ZAQAR_MESSAGE
    ) = (
        'POLL_SERVER_CFN', 'POLL_SERVER_HEAT', 'POLL_TEMP_URL', 'ZAQAR_MESSAGE'
    )

    _DEPLOYMENT_SWIFT_DATA_KEYS = (
        CONTAINER, OBJECT
    ) = (
        'container', 'object',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Server name.'),
            update_allowed=True
        ),
        METADATA: properties.Schema(
            properties.Schema.MAP,
            _('Arbitrary key/value metadata to store for this server. Both '
              'keys and values must be 255 characters or less. Non-string '
              'values will be serialized to JSON (and the serialized '
              'string must be 255 characters or less).'),
            update_allowed=True,
            support_status=support.SupportStatus(
                status=support.DEPRECATED,
                message='This property will be ignored',
                version='9.0.0',
                previous_status=support.SupportStatus(
                    status=support.SUPPORTED,
                    version='8.0.0'
                )
            )
        ),
        SOFTWARE_CONFIG_TRANSPORT: properties.Schema(
            properties.Schema.STRING,
            _('How the server should receive the metadata required for '
              'software configuration. POLL_SERVER_CFN will allow calls to '
              'the cfn API action DescribeStackResource authenticated with '
              'the provided keypair. POLL_SERVER_HEAT will allow calls to '
              'the Heat API resource-show using the provided keystone '
              'credentials. POLL_TEMP_URL will create and populate a '
              'Swift TempURL with metadata for polling. ZAQAR_MESSAGE will '
              'create a dedicated zaqar queue and post the metadata '
              'for polling.'),
            default=cfg.CONF.default_software_config_transport,
            update_allowed=True,
            constraints=[
                constraints.AllowedValues(_SOFTWARE_CONFIG_TRANSPORTS),
            ]
        ),
        DEPLOYMENT_SWIFT_DATA: properties.Schema(
            properties.Schema.MAP,
            _('Swift container and object to use for storing deployment data '
              'for the server resource. The parameter is a map value '
              'with the keys "container" and "object", and the values '
              'are the corresponding container and object names. The '
              'software_config_transport parameter must be set to '
              'POLL_TEMP_URL for swift to be used. If not specified, '
              'and software_config_transport is set to POLL_TEMP_URL, a '
              'container will be automatically created from the resource '
              'name, and the object name will be a generated uuid.'),
            support_status=support.SupportStatus(version='9.0.0'),
            default={},
            update_allowed=True,
            schema={
                CONTAINER: properties.Schema(
                    properties.Schema.STRING,
                    _('Name of the container.'),
                    constraints=[
                        constraints.Length(min=1)
                    ]
                ),
                OBJECT: properties.Schema(
                    properties.Schema.STRING,
                    _('Name of the object.'),
                    constraints=[
                        constraints.Length(min=1)
                    ]
                )
            }
        )
    }

    ATTRIBUTES = (
        NAME_ATTR, OS_COLLECT_CONFIG
    ) = (
        'name', 'os_collect_config'
    )

    attributes_schema = {
        NAME_ATTR: attributes.Schema(
            _('Name of the server.'),
            type=attributes.Schema.STRING
        ),
        OS_COLLECT_CONFIG: attributes.Schema(
            _('The os-collect-config configuration for the server\'s local '
              'agent to be configured to connect to Heat to retrieve '
              'deployment data.'),
            type=attributes.Schema.MAP,
            support_status=support.SupportStatus(version='9.0.0'),
            cache_mode=attributes.Schema.CACHE_NONE
        ),
    }

    default_client_name = 'heat'

    def __init__(self, name, json_snippet, stack):
        super(DeployedServer, self).__init__(name, json_snippet, stack)
        self._register_access_key()

    def handle_create(self):
        metadata = self.metadata_get(True) or {}
        self.resource_id_set(self.uuid)

        self._create_transport_credentials(self.properties)
        self._populate_deployments_metadata(metadata, self.properties)

        return self.resource_id

    def user_data_software_config(self):
        return True

    def _delete(self):
        self._delete_queue()
        self._delete_user()
        self._delete_temp_url()


def resource_mapping():
    return {
        'OS::Heat::DeployedServer': DeployedServer,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from webob import exc

from heat.api.openstack.v1 import util
from heat.common.i18n import _
from heat.common import identifier
from heat.common import param_utils
from heat.common import serializers
from heat.common import wsgi
from heat.rpc import api as rpc_api
from heat.rpc import client as rpc_client


summary_keys = [
    rpc_api.EVENT_ID,
    rpc_api.EVENT_TIMESTAMP,
    rpc_api.EVENT_RES_NAME,
    rpc_api.EVENT_RES_STATUS,
    rpc_api.EVENT_RES_STATUS_DATA,
    rpc_api.EVENT_RES_PHYSICAL_ID,
]


def format_event(req, event, keys=None):

    def include_key(k):
        return k in keys if keys else True

    def transform(key, value):
        if not include_key(key):
            return

        if key == rpc_api.EVENT_ID:
            identity = identifier.EventIdentifier(**value)
            yield ('id', identity.event_id)
            yield ('links', [util.make_link(req, identity),
                             util.make_link(req, identity.resource(),
                                            'resource'),
                             util.make_link(req, identity.stack(),
                                            'stack')])
        elif key in (rpc_api.EVENT_STACK_ID, rpc_api.EVENT_STACK_NAME,
                     rpc_api.EVENT_RES_ACTION):
            return
        elif (key == rpc_api.EVENT_RES_STATUS and
              rpc_api.EVENT_RES_ACTION in event):
            # To avoid breaking API compatibility, we join RES_ACTION
            # and RES_STATUS, so the API format doesn't expose the
            # internal split of state into action/status
            yield (key, '_'.join((event[rpc_api.EVENT_RES_ACTION], value)))
        elif (key == rpc_api.RES_NAME):
            yield ('logical_resource_id', value)
            yield (key, value)

        else:
            yield (key, value)

    ev = dict(itertools.chain.from_iterable(
        transform(k, v) for k, v in event.items()))

    root_stack_id = event.get(rpc_api.EVENT_ROOT_STACK_ID)
    if root_stack_id:
        root_identifier = identifier.HeatIdentifier(**root_stack_id)
        ev['links'].append(util.make_link(req, root_identifier, 'root_stack'))
    return ev


class EventController(object):
    """WSGI controller for Events in Heat v1 API.

    Implements the API actions.
    """
    # Define request scope (must match what is in policy.json or policies in
    # code)
    REQUEST_SCOPE = 'events'

    def __init__(self, options):
        self.options = options
        self.rpc_client = rpc_client.EngineClient()

    def _event_list(self, req, identity, detail=False, filters=None,
                    limit=None, marker=None, sort_keys=None, sort_dir=None,
                    nested_depth=None):
        events = self.rpc_client.list_events(req.context,
                                             identity,
                                             filters=filters,
                                             limit=limit,
                                             marker=marker,
                                             sort_keys=sort_keys,
                                             sort_dir=sort_dir,
                                             nested_depth=nested_depth)
        keys = None if detail else summary_keys

        return [format_event(req, e, keys) for e in events]

    @util.registered_identified_stack
    def index(self, req, identity, resource_name=None):
        """Lists summary information for all events."""
        param_types = {
            'limit': util.PARAM_TYPE_SINGLE,
            'marker': util.PARAM_TYPE_SINGLE,
            'sort_dir': util.PARAM_TYPE_SINGLE,
            'sort_keys': util.PARAM_TYPE_MULTI,
            'nested_depth': util.PARAM_TYPE_SINGLE,
        }
        filter_param_types = {
            'resource_status': util.PARAM_TYPE_MIXED,
            'resource_action': util.PARAM_TYPE_MIXED,
            'resource_name': util.PARAM_TYPE_MIXED,
            'resource_type': util.PARAM_TYPE_MIXED,
        }
        params = util.get_allowed_params(req.params, param_types)
        filter_params = util.get_allowed_params(req.params, filter_param_types)

        int_params = (rpc_api.PARAM_LIMIT, rpc_api.PARAM_NESTED_DEPTH)
        try:
            for key in int_params:
                if key in params:
                    params[key] = param_utils.extract_int(
                        key, params[key], allow_zero=True)
        except ValueError as e:
            raise exc.HTTPBadRequest(str(e))

        if resource_name is None:
            if not filter_params:
                filter_params = None
        else:
            filter_params['resource_name'] = resource_name

        events = self._event_list(
            req, identity, filters=filter_params, **params)

        if not events and resource_name is not None:
            msg = _('No events found for resource %s') % resource_name
            raise exc.HTTPNotFound(msg)

        return {'events': events}

    @util.registered_identified_stack
    def show(self, req, identity, resource_name, event_id):
        """Gets detailed information for an event."""

        filters = {"resource_name": resource_name, "uuid": event_id}
        events = self._event_list(req, identity, filters=filters, detail=True)
        if not events:
            raise exc.HTTPNotFound(_('No event %s found') % event_id)

        return {'event': events[0]}


def create_resource(options):
    """Events resource factory method."""
    deserializer = wsgi.JSONRequestDeserializer()
    serializer = serializers.JSONResponseSerializer()
    return wsgi.Resource(EventController(options), deserializer, serializer)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import status
from heat.engine import template
from heat.rpc import api as rpc_api


class GroupInspector(object):
    """A class for returning data about a scaling group.

    All data is fetched over RPC, and the group's stack is never loaded into
    memory locally. Data is cached so it will be fetched only once. To
    refresh the data, create a new GroupInspector.
    """

    def __init__(self, context, rpc_client, group_identity):
        """Initialise with a context, rpc_client, and stack identifier."""
        self._context = context
        self._rpc_client = rpc_client
        self._identity = group_identity
        self._member_data = None
        self._template_data = None

    @classmethod
    def from_parent_resource(cls, parent_resource):
        """Create a GroupInspector from a parent resource.

        This is a convenience method to instantiate a GroupInspector from a
        Heat StackResource object.
        """
        return cls(parent_resource.context, parent_resource.rpc_client(),
                   parent_resource.nested_identifier())

    def _get_member_data(self):
        if self._identity is None:
            return []

        if self._member_data is None:
            rsrcs = self._rpc_client.list_stack_resources(self._context,
                                                          dict(self._identity))

            def sort_key(r):
                return (r[rpc_api.RES_STATUS] != status.ResourceStatus.FAILED,
                        r[rpc_api.RES_CREATION_TIME],
                        r[rpc_api.RES_NAME])

            self._member_data = sorted(rsrcs, key=sort_key)

        return self._member_data

    def _members(self, include_failed):
        return (r for r in self._get_member_data()
                if (include_failed or
                    r[rpc_api.RES_STATUS] != status.ResourceStatus.FAILED))

    def size(self, include_failed):
        """Return the size of the group.

        If include_failed is False, only members not in a FAILED state will
        be counted.
        """
        return sum(1 for m in self._members(include_failed))

    def member_names(self, include_failed):
        """Return an iterator over the names of the group members

        If include_failed is False, only members not in a FAILED state will
        be included.
        """
        return (m[rpc_api.RES_NAME] for m in self._members(include_failed))

    def _get_template_data(self):
        if self._identity is None:
            return None

        if self._template_data is None:
            self._template_data = self._rpc_client.get_template(self._context,
                                                                self._identity)
        return self._template_data

    def template(self):
        """Return a Template object representing the group's current template.

        Note that this does *not* include any environment data.
        """
        data = self._get_template_data()
        if data is None:
            return None
        return template.Template(data)


def get_size(group, include_failed=False):
    """Get number of member resources managed by the specified group.

    The size excludes failed members by default; set include_failed=True
    to get the total size.
    """
    return GroupInspector.from_parent_resource(group).size(include_failed)


def get_members(group, include_failed=False):
    """Get a list of member resources managed by the specified group.

    Sort the list of instances first by created_time then by name.
    If include_failed is set, failed members will be put first in the
    list sorted by created_time then by name.
    """
    resources = []
    if group.nested():
        resources = [r for r in group.nested().values()
                     if include_failed or r.status != r.FAILED]

    return sorted(resources,
                  key=lambda r: (r.status != r.FAILED, r.created_time, r.name))


def get_member_refids(group):
    """Get a list of member resources managed by the specified group.

    The list of resources is sorted first by created_time then by name.
    """
    return [r.FnGetRefId() for r in get_members(group)]


def get_member_names(group):
    """Get a list of resource names of the resources in the specified group.

    Failed resources will be ignored.
    """
    inspector = GroupInspector.from_parent_resource(group)
    return list(inspector.member_names(include_failed=False))


def get_resource(stack, resource_name, use_indices, key=None):
    nested_stack = stack.nested()
    if not nested_stack:
        return None
    try:
        if use_indices:
            return get_members(stack)[int(resource_name)]
        else:
            return nested_stack[resource_name]
    except (IndexError, KeyError):
        raise exception.NotFound(_("Member '%(mem)s' not found "
                                   "in group resource '%(grp)s'.")
                                 % {'mem': resource_name,
                                    'grp': stack.name})


def get_rsrc_attr(stack, key, use_indices, resource_name, *attr_path):
    resource = get_resource(stack, resource_name, use_indices)
    if resource:
        return resource.FnGetAtt(*attr_path)


def get_rsrc_id(stack, key, use_indices, resource_name):
    resource = get_resource(stack, resource_name, use_indices)
    if resource:
        return resource.FnGetRefId()


def get_nested_attrs(stack, key, use_indices, *path):
    path = key.split(".", 2)[1:] + list(path)
    if len(path) > 1:
        return get_rsrc_attr(stack, key, use_indices, *path)
    else:
        return get_rsrc_id(stack, key, use_indices, *path)


def get_member_definitions(group, include_failed=False):
    """Get member definitions in (name, ResourceDefinition) pair for group.

    The List is sorted first by created_time then by name.
    If include_failed is set, failed members will be put first in the
    List sorted by created_time then by name.
    """
    inspector = GroupInspector.from_parent_resource(group)
    template = inspector.template()
    if template is None:
        return []
    definitions = template.resource_definitions(None)
    return [(name, definitions[name])
            for name in inspector.member_names(include_failed=include_failed)
            if name in definitions]


def get_child_template_files(context, stack,
                             is_rolling_update,
                             old_template_id):
    """Return a merged map of old and new template files.

    For rolling update files for old and new defintions are required as the
    nested stack is updated in batches of scaled units.
    """
    if not stack.convergence:
        old_template_id = stack.t.id

    if is_rolling_update and old_template_id:
        prev_files = template.Template.load(context, old_template_id).files
        prev_files.update(dict(stack.t.files))
        return prev_files
    return stack.t.files



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import excutils

from heat.common import exception
from heat.common import grouputils
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import function
from heat.engine.notification import autoscaling as notification
from heat.engine import properties
from heat.engine import resource
from heat.engine.resources.openstack.heat import instance_group as instgrp
from heat.engine import rsrc_defn
from heat.engine import support
from heat.scaling import cooldown
from heat.scaling import scalingutil as sc_util

LOG = logging.getLogger(__name__)


class AutoScalingGroup(cooldown.CooldownMixin, instgrp.InstanceGroup):

    support_status = support.SupportStatus(version='2014.1')

    PROPERTIES = (
        AVAILABILITY_ZONES, LAUNCH_CONFIGURATION_NAME, MAX_SIZE, MIN_SIZE,
        COOLDOWN, DESIRED_CAPACITY, HEALTH_CHECK_GRACE_PERIOD,
        HEALTH_CHECK_TYPE, LOAD_BALANCER_NAMES, VPCZONE_IDENTIFIER, TAGS,
        INSTANCE_ID,
    ) = (
        'AvailabilityZones', 'LaunchConfigurationName', 'MaxSize', 'MinSize',
        'Cooldown', 'DesiredCapacity', 'HealthCheckGracePeriod',
        'HealthCheckType', 'LoadBalancerNames', 'VPCZoneIdentifier', 'Tags',
        'InstanceId',
    )

    _TAG_KEYS = (
        TAG_KEY, TAG_VALUE,
    ) = (
        'Key', 'Value',
    )

    _UPDATE_POLICY_SCHEMA_KEYS = (
        ROLLING_UPDATE
    ) = (
        'AutoScalingRollingUpdate'
    )

    _ROLLING_UPDATE_SCHEMA_KEYS = (
        MIN_INSTANCES_IN_SERVICE, MAX_BATCH_SIZE, PAUSE_TIME
    ) = (
        'MinInstancesInService', 'MaxBatchSize', 'PauseTime'
    )

    ATTRIBUTES = (
        INSTANCE_LIST,
    ) = (
        'InstanceList',
    )

    properties_schema = {
        AVAILABILITY_ZONES: properties.Schema(
            properties.Schema.LIST,
            _('Not Implemented.'),
            required=True
        ),
        LAUNCH_CONFIGURATION_NAME: properties.Schema(
            properties.Schema.STRING,
            _('The reference to a LaunchConfiguration resource.'),
            update_allowed=True
        ),
        INSTANCE_ID: properties.Schema(
            properties.Schema.STRING,
            _('The ID of an existing instance to use to '
              'create the Auto Scaling group. If specify this property, '
              'will create the group use an existing instance instead of '
              'a launch configuration.'),
            constraints=[
                constraints.CustomConstraint("nova.server")
            ]
        ),
        MAX_SIZE: properties.Schema(
            properties.Schema.INTEGER,
            _('Maximum number of instances in the group.'),
            required=True,
            update_allowed=True
        ),
        MIN_SIZE: properties.Schema(
            properties.Schema.INTEGER,
            _('Minimum number of instances in the group.'),
            required=True,
            update_allowed=True
        ),
        COOLDOWN: properties.Schema(
            properties.Schema.INTEGER,
            _('Cooldown period, in seconds.'),
            update_allowed=True
        ),
        DESIRED_CAPACITY: properties.Schema(
            properties.Schema.INTEGER,
            _('Desired initial number of instances.'),
            update_allowed=True
        ),
        HEALTH_CHECK_GRACE_PERIOD: properties.Schema(
            properties.Schema.INTEGER,
            _('Not Implemented.'),
            implemented=False
        ),
        HEALTH_CHECK_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Not Implemented.'),
            constraints=[
                constraints.AllowedValues(['EC2', 'ELB']),
            ],
            implemented=False
        ),
        LOAD_BALANCER_NAMES: properties.Schema(
            properties.Schema.LIST,
            _('List of LoadBalancer resources.')
        ),
        VPCZONE_IDENTIFIER: properties.Schema(
            properties.Schema.LIST,
            _('Use only with Neutron, to list the internal subnet to '
              'which the instance will be attached; '
              'needed only if multiple exist; '
              'list length must be exactly 1.'),
            schema=properties.Schema(
                properties.Schema.STRING,
                _('UUID of the internal subnet to which the instance '
                  'will be attached.')
            )
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('Tags to attach to this group.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    TAG_KEY: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                    TAG_VALUE: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                },
            )
        ),
    }

    attributes_schema = {
        INSTANCE_LIST: attributes.Schema(
            _("A comma-delimited list of server ip addresses. "
              "(Heat extension)."),
            type=attributes.Schema.STRING
        ),
    }

    rolling_update_schema = {
        MIN_INSTANCES_IN_SERVICE: properties.Schema(properties.Schema.INTEGER,
                                                    default=0),
        MAX_BATCH_SIZE: properties.Schema(properties.Schema.INTEGER,
                                          default=1),
        PAUSE_TIME: properties.Schema(properties.Schema.STRING,
                                      default='PT0S')
    }

    update_policy_schema = {
        ROLLING_UPDATE: properties.Schema(properties.Schema.MAP,
                                          schema=rolling_update_schema)
    }

    def get_size(self):
        """Get desired capacity."""
        return self.properties[self.DESIRED_CAPACITY]

    def handle_create(self):
        return self.create_with_template(self.child_template())

    def _make_launch_config_resource(self, name, props):
        lc_res_type = 'AWS::AutoScaling::LaunchConfiguration'
        lc_res_def = rsrc_defn.ResourceDefinition(name,
                                                  lc_res_type,
                                                  props)
        lc_res = resource.Resource(name, lc_res_def, self.stack)
        return lc_res

    def _get_conf_properties(self):
        instance_id = self.properties.get(self.INSTANCE_ID)
        if instance_id:
            server = self.client_plugin('nova').get_server(instance_id)
            instance_props = {
                'ImageId': server.image['id'],
                'InstanceType': server.flavor['id'],
                'KeyName': server.key_name,
                'SecurityGroups': [sg['name']
                                   for sg in server.security_groups]
            }
            conf = self._make_launch_config_resource(self.name,
                                                     instance_props)
            props = function.resolve(conf.properties.data)
        else:
            conf, props = super(AutoScalingGroup, self)._get_conf_properties()

        vpc_zone_ids = self.properties.get(self.VPCZONE_IDENTIFIER)
        if vpc_zone_ids:
            props['SubnetId'] = vpc_zone_ids[0]

        return conf, props

    def check_create_complete(self, task):
        """Update cooldown timestamp after create succeeds."""
        done = super(AutoScalingGroup, self).check_create_complete(task)
        cooldown = self.properties[self.COOLDOWN]
        if done:
            self._finished_scaling(cooldown,
                                   "%s : %s" % (sc_util.CFN_EXACT_CAPACITY,
                                                grouputils.get_size(self)))
        return done

    def check_update_complete(self, cookie):
        """Update the cooldown timestamp after update succeeds."""
        done = super(AutoScalingGroup, self).check_update_complete(cookie)
        cooldown = self.properties[self.COOLDOWN]
        if done:
            self._finished_scaling(cooldown,
                                   "%s : %s" % (sc_util.CFN_EXACT_CAPACITY,
                                                grouputils.get_size(self)))
        return done

    def _get_new_capacity(self, capacity,
                          adjustment,
                          adjustment_type=sc_util.CFN_EXACT_CAPACITY,
                          min_adjustment_step=None):
        lower = self.properties[self.MIN_SIZE]
        upper = self.properties[self.MAX_SIZE]
        return sc_util.calculate_new_capacity(capacity, adjustment,
                                              adjustment_type,
                                              min_adjustment_step,
                                              lower, upper)

    def resize(self, capacity):
        try:
            super(AutoScalingGroup, self).resize(capacity)
        finally:
            # allow InstanceList to be re-resolved
            self.clear_stored_attributes()

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        """Updates self.properties, if Properties has changed.

        If Properties has changed, update self.properties, so we get the new
        values during any subsequent adjustment.
        """
        if tmpl_diff:
            # parse update policy
            if tmpl_diff.update_policy_changed():
                up = json_snippet.update_policy(self.update_policy_schema,
                                                self.context)
                self.update_policy = up

        self.properties = json_snippet.properties(self.properties_schema,
                                                  self.context)
        if prop_diff:
            # Replace instances first if launch configuration has changed
            self._try_rolling_update(prop_diff)

        # Update will happen irrespective of whether auto-scaling
        # is in progress or not.
        capacity = grouputils.get_size(self)
        desired_capacity = self.properties[self.DESIRED_CAPACITY] or capacity
        new_capacity = self._get_new_capacity(capacity, desired_capacity)
        self.resize(new_capacity)

    def adjust(self, adjustment,
               adjustment_type=sc_util.CFN_CHANGE_IN_CAPACITY,
               min_adjustment_step=None, cooldown=None):
        """Adjust the size of the scaling group if the cooldown permits."""
        if self.status != self.COMPLETE:
            LOG.info("%s NOT performing scaling adjustment, "
                     "when status is not COMPLETE", self.name)
            raise resource.NoActionRequired

        capacity = grouputils.get_size(self)
        new_capacity = self._get_new_capacity(capacity, adjustment,
                                              adjustment_type,
                                              min_adjustment_step)
        if new_capacity == capacity:
            LOG.info("%s NOT performing scaling adjustment, "
                     "as there is no change in capacity.", self.name)
            raise resource.NoActionRequired

        if cooldown is None:
            cooldown = self.properties[self.COOLDOWN]

        self._check_scaling_allowed(cooldown)

        # send a notification before, on-error and on-success.
        notif = {
            'stack': self.stack,
            'adjustment': adjustment,
            'adjustment_type': adjustment_type,
            'capacity': capacity,
            'groupname': self.FnGetRefId(),
            'message': _("Start resizing the group %(group)s") % {
                'group': self.FnGetRefId()},
            'suffix': 'start',
        }
        size_changed = False
        try:
            notification.send(**notif)
            try:
                self.resize(new_capacity)
            except Exception as resize_ex:
                with excutils.save_and_reraise_exception():
                    try:
                        notif.update({'suffix': 'error',
                                      'message': str(resize_ex),
                                      'capacity': grouputils.get_size(self),
                                      })
                        notification.send(**notif)
                    except Exception:
                        LOG.exception('Failed sending error notification')
            else:
                size_changed = True
                notif.update({
                    'suffix': 'end',
                    'capacity': new_capacity,
                    'message': _("End resizing the group %(group)s") % {
                        'group': notif['groupname']},
                })
                notification.send(**notif)
        except Exception:
            LOG.error("Error in performing scaling adjustment for "
                      "group %s.", self.name)
            raise
        finally:
            self._finished_scaling(cooldown,
                                   "%s : %s" % (adjustment_type, adjustment),
                                   size_changed=size_changed)

    def _tags(self):
        """Add Identifying Tags to all servers in the group.

        This is so the Dimensions received from cfn-push-stats all include
        the groupname and stack id.
        Note: the group name must match what is returned from FnGetRefId
        """
        autoscaling_tag = [{self.TAG_KEY: 'metering.AutoScalingGroupName',
                            self.TAG_VALUE: self.FnGetRefId()}]
        return super(AutoScalingGroup, self)._tags() + autoscaling_tag

    def validate(self):
        # check validity of group size
        min_size = self.properties[self.MIN_SIZE]
        max_size = self.properties[self.MAX_SIZE]

        if max_size < min_size:
            msg = _("MinSize can not be greater than MaxSize")
            raise exception.StackValidationFailed(message=msg)

        if min_size < 0:
            msg = _("The size of AutoScalingGroup can not be less than zero")
            raise exception.StackValidationFailed(message=msg)

        if self.properties[self.DESIRED_CAPACITY] is not None:
            desired_capacity = self.properties[self.DESIRED_CAPACITY]
            if desired_capacity < min_size or desired_capacity > max_size:
                msg = _("DesiredCapacity must be between MinSize and MaxSize")
                raise exception.StackValidationFailed(message=msg)

        # TODO(pasquier-s): once Neutron is able to assign subnets to
        # availability zones, it will be possible to specify multiple subnets.
        # For now, only one subnet can be specified. The bug #1096017 tracks
        # this issue.
        if (self.properties.get(self.VPCZONE_IDENTIFIER) and
                len(self.properties[self.VPCZONE_IDENTIFIER]) != 1):
            raise exception.NotSupported(feature=_("Anything other than one "
                                         "VPCZoneIdentifier"))
        # validate properties InstanceId and LaunchConfigurationName
        # for aws auto scaling group.
        # should provide just only one of
        if self.type() == 'AWS::AutoScaling::AutoScalingGroup':
            instanceId = self.properties.get(self.INSTANCE_ID)
            launch_config = self.properties.get(
                self.LAUNCH_CONFIGURATION_NAME)
            if bool(instanceId) == bool(launch_config):
                msg = _("Either 'InstanceId' or 'LaunchConfigurationName' "
                        "must be provided.")
                raise exception.StackValidationFailed(message=msg)

        super(AutoScalingGroup, self).validate()

    def child_template(self):
        if self.properties[self.DESIRED_CAPACITY]:
            num_instances = self.properties[self.DESIRED_CAPACITY]
        else:
            num_instances = self.properties[self.MIN_SIZE]
        return self._create_template(num_instances)


def resource_mapping():
    return {
        'AWS::AutoScaling::AutoScalingGroup': AutoScalingGroup,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.resources.openstack.barbican import order
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils

stack_template = '''
heat_template_version: 2013-05-23
description: Test template
resources:
  order:
    type: OS::Barbican::Order
    properties:
      name: foobar-order
      algorithm: aes
      bit_length: 256
      mode: cbc
      type: key
'''


class FakeOrder(object):

    def __init__(self, name):
        self.name = name

    def submit(self):
        return self.name


class TestOrder(common.HeatTestCase):

    def setUp(self):
        super(TestOrder, self).setUp()
        tmpl = template_format.parse(stack_template)
        self.stack = utils.parse_stack(tmpl)

        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.res_template = resource_defns['order']
        self.props = tmpl['resources']['order']['properties']

        self.patcher_client = mock.patch.object(order.Order, 'client')
        mock_client = self.patcher_client.start()
        self.barbican = mock_client.return_value

    def tearDown(self):
        super(TestOrder, self).tearDown()
        self.patcher_client.stop()

    def _create_resource(self, name, snippet, stack):
        res = order.Order(name, snippet, stack)
        res.check_create_complete = mock.Mock(return_value=True)
        self.barbican.orders.create.return_value = FakeOrder(name)
        scheduler.TaskRunner(res.create)()
        return res

    def test_create_order(self):
        res = self._create_resource('foo', self.res_template, self.stack)
        expected_state = (res.CREATE, res.COMPLETE)
        self.assertEqual(expected_state, res.state)
        args = self.barbican.orders.create.call_args[1]
        self.assertEqual('foobar-order', args['name'])
        self.assertEqual('aes', args['algorithm'])
        self.assertEqual('cbc', args['mode'])
        self.assertEqual(256, args['bit_length'])

    def test_create_order_without_type_fail(self):
        props = self.props.copy()
        del props['type']
        snippet = self.res_template.freeze(properties=props)

        self.assertRaisesRegex(exception.ResourceFailure,
                               'Property type not assigned',
                               self._create_resource,
                               'foo',
                               snippet, self.stack)

    def test_validate_non_certificate_order(self):
        props = self.props.copy()
        del props['bit_length']
        del props['algorithm']
        snippet = self.res_template.freeze(properties=props)
        res = self._create_resource('test', snippet, self.stack)
        msg = ("Properties algorithm and bit_length are required for "
               "key type of order.")
        self.assertRaisesRegex(exception.StackValidationFailed,
                               msg,
                               res.validate)

    def test_validate_certificate_with_profile_without_ca_id(self):
        props = self.props.copy()
        props['profile'] = 'cert'
        props['type'] = 'certificate'
        snippet = self.res_template.freeze(properties=props)
        res = self._create_resource('test', snippet, self.stack)
        msg = ("profile cannot be specified without ca_id.")
        self.assertRaisesRegex(exception.ResourcePropertyDependency,
                               msg,
                               res.validate)

    def test_key_order_validation_fail(self):
        props = self.props.copy()
        props['pass_phrase'] = "something"
        snippet = self.res_template.freeze(properties=props)
        res = self._create_resource('test', snippet, self.stack)
        msg = ("Unexpected properties: pass_phrase. Only these properties "
               "are allowed for key type of order: algorithm, "
               "bit_length, expiration, mode, name, payload_content_type.")
        self.assertRaisesRegex(exception.StackValidationFailed,
                               msg,
                               res.validate)

    def test_certificate_validation_fail(self):
        props = self.props.copy()
        props['type'] = 'certificate'
        snippet = self.res_template.freeze(properties=props)
        res = self._create_resource('test', snippet, self.stack)
        msg = ("Unexpected properties: algorithm, bit_length, mode. Only "
               "these properties are allowed for certificate type of order: "
               "ca_id, name, profile, request_data, request_type, "
               "source_container_ref, subject_dn.")
        self.assertRaisesRegex(exception.StackValidationFailed,
                               msg,
                               res.validate)

    def test_asymmetric_order_validation_fail(self):
        props = self.props.copy()
        props['type'] = 'asymmetric'
        props['subject_dn'] = 'asymmetric'
        snippet = self.res_template.freeze(properties=props)
        res = self._create_resource('test', snippet, self.stack)
        msg = ("Unexpected properties: subject_dn. Only these properties are "
               "allowed for asymmetric type of order: algorithm, bit_length, "
               "expiration, mode, name, pass_phrase, payload_content_type")
        self.assertRaisesRegex(exception.StackValidationFailed,
                               msg,
                               res.validate)

    def test_attributes(self):
        mock_order = mock.Mock()
        mock_order.status = 'test-status'
        mock_order.order_ref = 'test-order-ref'
        mock_order.secret_ref = 'test-secret-ref'

        res = self._create_resource('foo', self.res_template, self.stack)
        self.barbican.orders.get.return_value = mock_order

        self.assertEqual('test-order-ref', res.FnGetAtt('order_ref'))
        self.assertEqual('test-secret-ref', res.FnGetAtt('secret_ref'))

    def test_attributes_handle_exceptions(self):
        mock_order = mock.Mock()
        res = self._create_resource('foo', self.res_template, self.stack)
        self.barbican.orders.get.return_value = mock_order

        self.barbican.barbican_client.HTTPClientError = Exception
        self.barbican.orders.get.side_effect = Exception('boom')
        self.assertRaises(self.barbican.barbican_client.HTTPClientError,
                          res.FnGetAtt, 'order_ref')

    def test_container_attributes(self):
        mock_order = mock.Mock()
        mock_order.container_ref = 'test-container-ref'

        mock_container = mock.Mock()
        mock_container.public_key = mock.Mock(payload='public-key')
        mock_container.private_key = mock.Mock(payload='private-key')
        mock_container.certificate = mock.Mock(payload='cert')
        mock_container.intermediates = mock.Mock(payload='interm')

        res = self._create_resource('foo', self.res_template, self.stack)
        self.barbican.orders.get.return_value = mock_order

        self.barbican.containers.get.return_value = mock_container

        self.assertEqual('public-key', res.FnGetAtt('public_key'))
        self.barbican.containers.get.assert_called_once_with(
            'test-container-ref')

        self.assertEqual('private-key', res.FnGetAtt('private_key'))
        self.assertEqual('cert', res.FnGetAtt('certificate'))
        self.assertEqual('interm', res.FnGetAtt('intermediates'))

    def test_create_order_sets_resource_id(self):
        self.barbican.orders.create.return_value = FakeOrder('foo')
        res = self._create_resource('foo', self.res_template, self.stack)

        self.assertEqual('foo', res.resource_id)

    def test_create_order_with_octet_stream(self):
        content_type = 'application/octet-stream'
        self.props['payload_content_type'] = content_type
        defn = rsrc_defn.ResourceDefinition('foo', 'OS::Barbican::Order',
                                            self.props)
        res = self._create_resource(defn.name, defn, self.stack)

        args = self.barbican.orders.create.call_args[1]
        self.assertEqual(content_type, args[res.PAYLOAD_CONTENT_TYPE])

    def test_check_create_complete(self):
        res = order.Order('foo', self.res_template, self.stack)

        mock_active = mock.Mock(status='ACTIVE')
        self.barbican.orders.get.return_value = mock_active
        self.assertTrue(res.check_create_complete('foo'))

        mock_not_active = mock.Mock(status='PENDING')
        self.barbican.orders.get.return_value = mock_not_active
        self.assertFalse(res.check_create_complete('foo'))

        mock_not_active = mock.Mock(status='ERROR', error_reason='foo',
                                    error_status_code=500)
        self.barbican.orders.get.return_value = mock_not_active
        exc = self.assertRaises(exception.Error,
                                res.check_create_complete, 'foo')
        self.assertIn('foo', str(exc))
        self.assertIn('500', str(exc))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class CinderVolumeType(resource.Resource):
    """A resource for creating cinder volume types.

    Volume type resource allows to define, whether volume, which will be use
    this type, will public and which projects are allowed to work with it.
    Also, there can be some user-defined metadata.

    Note that default cinder security policy usage of this resource
    is limited to being used by administrators only.
    """

    support_status = support.SupportStatus(version='2015.1')

    default_client_name = 'cinder'

    entity = 'volume_types'

    required_service_extension = 'os-types-manage'

    PROPERTIES = (
        NAME, METADATA, IS_PUBLIC, DESCRIPTION, PROJECTS,
    ) = (
        'name', 'metadata', 'is_public', 'description', 'projects',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the volume type.'),
            required=True,
            update_allowed=True,
        ),
        METADATA: properties.Schema(
            properties.Schema.MAP,
            _('The extra specs key and value pairs of the volume type.'),
            update_allowed=True
        ),
        IS_PUBLIC: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Whether the volume type is accessible to the public.'),
            default=True,
            support_status=support.SupportStatus(version='5.0.0'),
            update_allowed=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of the volume type.'),
            update_allowed=True,
            support_status=support.SupportStatus(version='5.0.0'),
        ),
        PROJECTS: properties.Schema(
            properties.Schema.LIST,
            _('Projects to add volume type access to. NOTE: This '
              'property is only supported since Cinder API V2.'),
            support_status=support.SupportStatus(version='5.0.0'),
            update_allowed=True,
            schema=properties.Schema(
                properties.Schema.STRING,
                constraints=[
                    constraints.CustomConstraint('keystone.project')
                ],
            ),
            default=[],
        ),
    }

    def _add_projects_access(self, projects):
        for project in projects:
            project_id = self.client_plugin('keystone').get_project_id(project)
            self.client().volume_type_access.add_project_access(
                self.resource_id, project_id)

    def handle_create(self):
        args = {
            'name': self.properties[self.NAME],
            'is_public': self.properties[self.IS_PUBLIC],
            'description': self.properties[self.DESCRIPTION]
        }

        volume_type = self.client().volume_types.create(**args)
        self.resource_id_set(volume_type.id)
        vtype_metadata = self.properties[self.METADATA]
        if vtype_metadata:
            volume_type.set_keys(vtype_metadata)
        self._add_projects_access(self.properties[self.PROJECTS])

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        """Update the name, description and metadata for volume type."""

        update_args = {}
        # Update the name, description, is_public of cinder volume type
        is_public = self.properties[self.IS_PUBLIC]
        if self.DESCRIPTION in prop_diff:
            update_args['description'] = prop_diff.get(self.DESCRIPTION)
        if self.NAME in prop_diff:
            update_args['name'] = prop_diff.get(self.NAME)
        if self.IS_PUBLIC in prop_diff:
            is_public = prop_diff.get(self.IS_PUBLIC)
            update_args['is_public'] = is_public
        if update_args:
            self.client().volume_types.update(self.resource_id, **update_args)
        # Update the key-value pairs of cinder volume type.
        if self.METADATA in prop_diff:
            volume_type = self.client().volume_types.get(self.resource_id)
            old_keys = volume_type.get_keys()
            volume_type.unset_keys(old_keys)
            new_keys = prop_diff.get(self.METADATA)
            if new_keys is not None:
                volume_type.set_keys(new_keys)
        # Update the projects access for volume type
        if self.PROJECTS in prop_diff and not is_public:
            old_access_list = self.client().volume_type_access.list(
                self.resource_id)
            old_projects = [ac.to_dict()['project_id'] for
                            ac in old_access_list]
            new_projects = prop_diff.get(self.PROJECTS)
            # first remove the old projects access
            for project_id in (set(old_projects) - set(new_projects)):
                self.client().volume_type_access.remove_project_access(
                    self.resource_id, project_id)
            # add the new projects access
            self._add_projects_access(set(new_projects) - set(old_projects))

    def validate(self):
        super(CinderVolumeType, self).validate()

        if self.properties[self.PROJECTS]:
            if self.properties[self.IS_PUBLIC]:
                msg = (_('Can not specify property "%s" '
                         'if the volume type is public.') % self.PROJECTS)
                raise exception.StackValidationFailed(message=msg)

    def get_live_resource_data(self):
        try:
            resource_object = self.client().volume_types.get(self.resource_id)
            resource_data = resource_object.to_dict()
        except Exception as ex:
            if self.client_plugin().is_not_found(ex):
                raise exception.EntityNotFound(entity='Resource',
                                               name=self.name)
            raise
        return resource_object, resource_data

    def parse_live_resource_data(self, resource_properties, resource_data):
        resource_reality = {}
        resource_object, resource_data = resource_data

        resource_reality.update({
            self.NAME: resource_data.get(self.NAME),
            self.DESCRIPTION: resource_data.get(self.DESCRIPTION)
        })

        metadata = resource_object.get_keys()
        resource_reality.update({self.METADATA: metadata or {}})

        is_public = resource_data.get(self.IS_PUBLIC)
        resource_reality.update({self.IS_PUBLIC: is_public})
        projects = []
        if not is_public:
            accesses = self.client().volume_type_access.list(self.resource_id)
            for access in accesses:
                projects.append(access.to_dict().get('project_id'))
        resource_reality.update({self.PROJECTS: projects})

        return resource_reality


def resource_mapping():
    return {
        'OS::Cinder::VolumeType': CinderVolumeType
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils

try:
    from neutronclient.v2_0 import client as neutronclient
except ImportError:
    neutronclient = None

test_template = {
    'heat_template_version': '2013-05-23',
    'resources': {
        'my_nic': {
            'type': 'AWS::EC2::NetworkInterface',
            'properties': {
                'SubnetId': 'ssss'
            }
        }
    }
}


class NetworkInterfaceTest(common.HeatTestCase):
    def setUp(self):
        super(NetworkInterfaceTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.m_ss = self.patchobject(neutronclient.Client, 'show_subnet')
        self.m_cp = self.patchobject(neutronclient.Client, 'create_port')
        self.m_dp = self.patchobject(neutronclient.Client, 'delete_port')
        self.m_up = self.patchobject(neutronclient.Client, 'update_port')

    def mock_show_subnet(self):
        self.m_ss.return_value = {
            'subnet': {
                'name': 'my_subnet',
                'network_id': 'nnnn',
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'allocation_pools': [{'start': '10.0.0.2',
                                      'end': '10.0.0.254'}],
                'gateway_ip': '10.0.0.1',
                'ip_version': 4,
                'cidr': '10.0.0.0/24',
                'id': 'ssss',
                'enable_dhcp': False,
            }}

    def mock_create_network_interface(self, stack_name='my_stack',
                                      resource_name='my_nic',
                                      security_groups=None):
        self.nic_name = utils.PhysName(stack_name, resource_name)
        self.port = {'network_id': 'nnnn',
                     'fixed_ips': [{
                         'subnet_id': u'ssss'
                     }],
                     'name': self.nic_name,
                     'admin_state_up': True}

        port_info = {
            'port': {
                'admin_state_up': True,
                'device_id': '',
                'device_owner': '',
                'fixed_ips': [
                    {
                        'ip_address': '10.0.0.100',
                        'subnet_id': 'ssss'
                    }
                ],
                'id': 'pppp',
                'mac_address': 'fa:16:3e:25:32:5d',
                'name': self.nic_name,
                'network_id': 'nnnn',
                'status': 'ACTIVE',
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f'
            }
        }

        if security_groups is not None:
            self.port['security_groups'] = security_groups
            port_info['security_groups'] = security_groups
        else:
            port_info['security_groups'] = ['default']

        self.m_cp.return_value = port_info

    def test_network_interface_create_update_delete(self):
        my_stack = utils.parse_stack(test_template,
                                     stack_name='test_nif_cud_stack')
        nic_rsrc = my_stack['my_nic']

        self.mock_show_subnet()
        self.stub_SubnetConstraint_validate()
        self.mock_create_network_interface(my_stack.name)

        update_props = {}
        update_sg_ids = ['0389f747-7785-4757-b7bb-2ab07e4b09c3']
        update_props['security_groups'] = update_sg_ids

        # create the nic without GroupSet
        self.assertIsNone(nic_rsrc.validate())
        scheduler.TaskRunner(nic_rsrc.create)()
        self.assertEqual((nic_rsrc.CREATE, my_stack.COMPLETE),
                         nic_rsrc.state)

        # update the nic with GroupSet
        props = copy.deepcopy(nic_rsrc.properties.data)
        props['GroupSet'] = update_sg_ids
        update_snippet = rsrc_defn.ResourceDefinition(nic_rsrc.name,
                                                      nic_rsrc.type(),
                                                      props)
        scheduler.TaskRunner(nic_rsrc.update, update_snippet)()
        self.assertEqual((nic_rsrc.UPDATE, nic_rsrc.COMPLETE), nic_rsrc.state)

        # delete the nic
        scheduler.TaskRunner(nic_rsrc.delete)()
        self.assertEqual((nic_rsrc.DELETE, nic_rsrc.COMPLETE), nic_rsrc.state)

        self.m_ss.assert_called_once_with('ssss')
        self.m_cp.assert_called_once_with({'port': self.port})
        self.m_up.assert_called_once_with('pppp', {'port': update_props})
        self.m_dp.assert_called_once_with('pppp')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import itertools

from heat.policies import actions
from heat.policies import base
from heat.policies import build_info
from heat.policies import cloudformation
from heat.policies import events
from heat.policies import resource
from heat.policies import resource_types
from heat.policies import service
from heat.policies import software_configs
from heat.policies import software_deployments
from heat.policies import stacks


def list_rules():
    return itertools.chain(
        base.list_rules(),
        actions.list_rules(),
        build_info.list_rules(),
        cloudformation.list_rules(),
        events.list_rules(),
        resource.list_rules(),
        resource_types.list_rules(),
        service.list_rules(),
        software_configs.list_rules(),
        software_deployments.list_rules(),
        stacks.list_rules(),
    )



# -*- coding: utf-8 -*-
#
# Copyright © 2013 Unitedstack Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A middleware that turns exceptions into parsable string.

Inspired by Cinder's faultwrapper.
"""

import sys
import traceback

from oslo_config import cfg
from oslo_utils import reflection
import webob

from heat.common import exception
from heat.common import serializers
from heat.common import wsgi


class Fault(object):

    def __init__(self, error):
        self.error = error

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        if req.content_type == 'application/xml':
            serializer = serializers.XMLResponseSerializer()
        else:
            serializer = serializers.JSONResponseSerializer()
        resp = webob.Response(request=req)
        default_webob_exc = webob.exc.HTTPInternalServerError()
        resp.status_code = self.error.get('code', default_webob_exc.code)
        serializer.default(resp, self.error)
        return resp


class FaultWrapper(wsgi.Middleware):
    """Replace error body with something the client can parse."""

    error_map = {
        'AttributeError': webob.exc.HTTPBadRequest,
        'ActionInProgress': webob.exc.HTTPConflict,
        'ValueError': webob.exc.HTTPBadRequest,
        'EntityNotFound': webob.exc.HTTPNotFound,
        'NotFound': webob.exc.HTTPNotFound,
        'ResourceActionNotSupported': webob.exc.HTTPBadRequest,
        'InvalidGlobalResource': webob.exc.HTTPInternalServerError,
        'ResourceNotAvailable': webob.exc.HTTPNotFound,
        'PhysicalResourceNameAmbiguity': webob.exc.HTTPBadRequest,
        'PhysicalResourceIDAmbiguity': webob.exc.HTTPBadRequest,
        'InvalidTenant': webob.exc.HTTPForbidden,
        'Forbidden': webob.exc.HTTPForbidden,
        'StackExists': webob.exc.HTTPConflict,
        'StackValidationFailed': webob.exc.HTTPBadRequest,
        'InvalidSchemaError': webob.exc.HTTPBadRequest,
        'InvalidTemplateReference': webob.exc.HTTPBadRequest,
        'InvalidTemplateVersion': webob.exc.HTTPBadRequest,
        'InvalidTemplateSection': webob.exc.HTTPBadRequest,
        'UnknownUserParameter': webob.exc.HTTPBadRequest,
        'RevertFailed': webob.exc.HTTPInternalServerError,
        'StopActionFailed': webob.exc.HTTPInternalServerError,
        'EventSendFailed': webob.exc.HTTPInternalServerError,
        'ServerBuildFailed': webob.exc.HTTPInternalServerError,
        'InvalidEncryptionKey': webob.exc.HTTPInternalServerError,
        'NotSupported': webob.exc.HTTPBadRequest,
        'MissingCredentialError': webob.exc.HTTPBadRequest,
        'UserParameterMissing': webob.exc.HTTPBadRequest,
        'RequestLimitExceeded': webob.exc.HTTPBadRequest,
        'DownloadLimitExceeded': webob.exc.HTTPBadRequest,
        'Invalid': webob.exc.HTTPBadRequest,
        'ResourcePropertyConflict': webob.exc.HTTPBadRequest,
        'PropertyUnspecifiedError': webob.exc.HTTPBadRequest,
        'ObjectFieldInvalid': webob.exc.HTTPBadRequest,
        'ReadOnlyFieldError': webob.exc.HTTPBadRequest,
        'ObjectActionError': webob.exc.HTTPBadRequest,
        'IncompatibleObjectVersion': webob.exc.HTTPBadRequest,
        'OrphanedObjectError': webob.exc.HTTPBadRequest,
        'UnsupportedObjectError': webob.exc.HTTPBadRequest,
        'ResourceTypeUnavailable': webob.exc.HTTPBadRequest,
        'InvalidBreakPointHook': webob.exc.HTTPBadRequest,
        'ImmutableParameterModified': webob.exc.HTTPBadRequest
    }

    def _map_exception_to_error(self, class_exception):
        if class_exception == Exception:
            return webob.exc.HTTPInternalServerError

        if class_exception.__name__ not in self.error_map:
            return self._map_exception_to_error(class_exception.__base__)

        return self.error_map[class_exception.__name__]

    def _error(self, ex):

        trace = None
        traceback_marker = 'Traceback (most recent call last)'
        webob_exc = None
        safe = getattr(ex, 'safe', False)

        if isinstance(ex, exception.HTTPExceptionDisguise):
            # An HTTP exception was disguised so it could make it here
            # let's remove the disguise and set the original HTTP exception
            if cfg.CONF.debug:
                trace = ''.join(traceback.format_tb(ex.tb))
            ex = ex.exc
            webob_exc = ex

        ex_type = reflection.get_class_name(ex, fully_qualified=False)

        is_remote = ex_type.endswith('_Remote')
        if is_remote:
            ex_type = ex_type[:-len('_Remote')]

        full_message = str(ex)
        if '\n' in full_message and is_remote:
            message, msg_trace = full_message.split('\n', 1)
        elif traceback_marker in full_message:
            message, msg_trace = full_message.split(traceback_marker, 1)
            message = message.rstrip('\n')
            msg_trace = traceback_marker + msg_trace
        else:
            msg_trace = 'None\n'
            if sys.exc_info() != (None, None, None):
                msg_trace = traceback.format_exc()
            message = full_message

        if isinstance(ex, exception.HeatException):
            message = ex.message

        if cfg.CONF.debug and not trace:
            trace = msg_trace

        if not webob_exc:
            webob_exc = self._map_exception_to_error(ex.__class__)

        error = {
            'code': webob_exc.code,
            'title': webob_exc.title,
            'explanation': webob_exc.explanation,
            'error': {
                'type': ex_type,
                'traceback': trace,
            }
        }
        if safe:
            error['error']['message'] = message

        return error

    def process_request(self, req):
        try:
            return req.get_response(self.application)
        except Exception as exc:
            return req.get_response(Fault(self._error(exc)))



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from openstack import exceptions

from heat.common import template_format
from heat.engine.clients.os import senlin
from heat.engine.resources.openstack.senlin import receiver as sr
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


receiver_stack_template = """
heat_template_version: 2016-04-08
description: Senlin Receiver Template
resources:
  senlin-receiver:
    type: OS::Senlin::Receiver
    properties:
      name: SenlinReceiver
      cluster: fake_cluster
      action: CLUSTER_SCALE_OUT
      type: webhook
      params:
        foo: bar
"""


class FakeReceiver(object):
    def __init__(self, id='some_id'):
        self.id = id
        self.name = "SenlinReceiver"
        self.cluster_id = "fake_cluster"
        self.action = "CLUSTER_SCALE_OUT"
        self.channel = {'alarm_url': "http://foo.bar/webhooks/fake_url"}

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
            'cluster_id': self.cluster_id,
            'action': self.action,
            'channel': self.channel,
            'actor': {'trust_id': ['fake_trust_id']}
        }


class SenlinReceiverTest(common.HeatTestCase):
    def setUp(self):
        super(SenlinReceiverTest, self).setUp()
        self.senlin_mock = mock.MagicMock()
        self.patchobject(sr.Receiver, 'client',
                         return_value=self.senlin_mock)
        self.patchobject(senlin.ClusterConstraint, 'validate',
                         return_value=True)
        self.fake_r = FakeReceiver()
        self.t = template_format.parse(receiver_stack_template)

    def _init_recv(self, template):
        self.stack = utils.parse_stack(template)
        recv = self.stack['senlin-receiver']
        return recv

    def _create_recv(self, template):
        recv = self._init_recv(template)
        self.senlin_mock.create_receiver.return_value = self.fake_r
        self.senlin_mock.get_receiver.return_value = self.fake_r
        scheduler.TaskRunner(recv.create)()
        self.assertEqual((recv.CREATE, recv.COMPLETE),
                         recv.state)
        self.assertEqual(self.fake_r.id, recv.resource_id)
        return recv

    def test_recv_create_success(self):
        self._create_recv(self.t)
        expect_kwargs = {
            'name': 'SenlinReceiver',
            'cluster_id': 'fake_cluster',
            'action': 'CLUSTER_SCALE_OUT',
            'type': 'webhook',
            'params': {'foo': 'bar'},
        }
        self.senlin_mock.create_receiver.assert_called_once_with(
            **expect_kwargs)

    def test_recv_delete_success(self):
        self.senlin_mock.delete_receiver.return_value = None
        recv = self._create_recv(self.t)
        scheduler.TaskRunner(recv.delete)()
        self.senlin_mock.delete_receiver.assert_called_once_with(
            recv.resource_id)

    def test_recv_delete_not_found(self):
        self.senlin_mock.delete_receiver.side_effect = [
            exceptions.ResourceNotFound(http_status=404)
        ]
        recv = self._create_recv(self.t)
        scheduler.TaskRunner(recv.delete)()
        self.senlin_mock.delete_receiver.assert_called_once_with(
            recv.resource_id)

    def test_cluster_resolve_attribute(self):
        excepted_show = {
            'id': 'some_id',
            'name': 'SenlinReceiver',
            'cluster_id': 'fake_cluster',
            'action': 'CLUSTER_SCALE_OUT',
            'channel': {'alarm_url': "http://foo.bar/webhooks/fake_url"},
            'actor': {'trust_id': ['fake_trust_id']}
        }
        recv = self._create_recv(self.t)
        self.assertEqual(self.fake_r.channel,
                         recv._resolve_attribute('channel'))
        self.assertEqual(excepted_show,
                         recv._show_resource())



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections


Message = collections.namedtuple('Message', ['name', 'data'])


class MessageQueue(object):

    def __init__(self, name):
        self.name = name
        self._queue = collections.deque()

    def send(self, name, data=None):
        self._queue.append(Message(name, data))

    def send_priority(self, name, data=None):
        self._queue.appendleft(Message(name, data))

    def get(self):
        try:
            return self._queue.popleft()
        except IndexError:
            return None

    def clear(self):
        self._queue.clear()



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.resources.openstack.sahara import data_source
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


data_source_template = """
heat_template_version: 2015-10-15
resources:
  data-source:
    type: OS::Sahara::DataSource
    properties:
      name: my-ds
      type: swift
      url: swift://container.sahara/text
      credentials:
          user: admin
          password: swordfish
"""


class SaharaDataSourceTest(common.HeatTestCase):
    def setUp(self):
        super(SaharaDataSourceTest, self).setUp()
        t = template_format.parse(data_source_template)
        self.stack = utils.parse_stack(t)
        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.rsrc_defn = resource_defns['data-source']
        self.client = mock.Mock()
        self.patchobject(data_source.DataSource, 'client',
                         return_value=self.client)

    def _create_resource(self, name, snippet, stack):
        ds = data_source.DataSource(name, snippet, stack)
        value = mock.MagicMock(id='12345')
        self.client.data_sources.create.return_value = value
        scheduler.TaskRunner(ds.create)()
        return ds

    def test_create(self):
        ds = self._create_resource('data-source', self.rsrc_defn, self.stack)
        args = self.client.data_sources.create.call_args[1]
        expected_args = {
            'name': 'my-ds',
            'description': '',
            'data_source_type': 'swift',
            'url': 'swift://container.sahara/text',
            'credential_user': 'admin',
            'credential_pass': 'swordfish'
        }
        self.assertEqual(expected_args, args)
        self.assertEqual('12345', ds.resource_id)
        expected_state = (ds.CREATE, ds.COMPLETE)
        self.assertEqual(expected_state, ds.state)

    def test_update(self):
        ds = self._create_resource('data-source', self.rsrc_defn,
                                   self.stack)
        props = self.stack.t.t['resources']['data-source']['properties'].copy()
        props['type'] = 'hdfs'
        props['url'] = 'my/path'
        self.rsrc_defn = self.rsrc_defn.freeze(properties=props)
        scheduler.TaskRunner(ds.update, self.rsrc_defn)()
        data = {
            'name': 'my-ds',
            'description': '',
            'type': 'hdfs',
            'url': 'my/path',
            'credentials': {
                'user': 'admin',
                'password': 'swordfish'
            }
        }
        self.client.data_sources.update.assert_called_once_with(
            '12345', data)
        self.assertEqual((ds.UPDATE, ds.COMPLETE), ds.state)

    def test_show_attribute(self):
        ds = self._create_resource('data-source', self.rsrc_defn, self.stack)
        value = mock.MagicMock()
        value.to_dict.return_value = {'ds': 'info'}
        self.client.data_sources.get.return_value = value
        self.assertEqual({'ds': 'info'}, ds.FnGetAtt('show'))

    def test_validate_password_without_user(self):
        props = self.stack.t.t['resources']['data-source']['properties'].copy()
        del props['credentials']['user']
        self.rsrc_defn = self.rsrc_defn.freeze(properties=props)
        ds = data_source.DataSource('data-source', self.rsrc_defn, self.stack)
        ex = self.assertRaises(exception.StackValidationFailed, ds.validate)
        error_msg = ('Property error: resources.data-source.properties.'
                     'credentials: Property user not assigned')
        self.assertEqual(error_msg, str(ex))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat_integrationtests.functional import functional_base
from heatclient import exc as heat_exceptions


class ImmutableParametersTest(functional_base.FunctionalTestsBase):

    template_param_has_no_immutable_field = '''
heat_template_version: 2014-10-16
parameters:
  param1:
    type: string
    default: default_value
outputs:
  param1_output:
    description: 'parameter 1 details'
    value: { get_param: param1 }
'''

    template_param_has_immutable_field = '''
heat_template_version: 2014-10-16
parameters:
  param1:
    type: string
    default: default_value
    immutable: false
outputs:
  param1_output:
    description: 'parameter 1 details'
    value: { get_param: param1 }
'''

    def test_no_immutable_param_field(self):
        param1_create_value = 'value1'
        create_parameters = {"param1": param1_create_value}

        stack_identifier = self.stack_create(
            template=self.template_param_has_no_immutable_field,
            parameters=create_parameters
        )
        stack = self.client.stacks.get(stack_identifier)

        # Verify the value of the parameter
        self.assertEqual(param1_create_value,
                         self._stack_output(stack, 'param1_output'))

        param1_update_value = 'value2'
        update_parameters = {"param1": param1_update_value}

        self.update_stack(
            stack_identifier,
            template=self.template_param_has_no_immutable_field,
            parameters=update_parameters)

        stack = self.client.stacks.get(stack_identifier)

        # Verify the value of the updated parameter
        self.assertEqual(param1_update_value,
                         self._stack_output(stack, 'param1_output'))

    def test_immutable_param_field_allowed(self):
        param1_create_value = 'value1'
        create_parameters = {"param1": param1_create_value}

        stack_identifier = self.stack_create(
            template=self.template_param_has_immutable_field,
            parameters=create_parameters
        )
        stack = self.client.stacks.get(stack_identifier)

        # Verify the value of the parameter
        self.assertEqual(param1_create_value,
                         self._stack_output(stack, 'param1_output'))

        param1_update_value = 'value2'
        update_parameters = {"param1": param1_update_value}

        self.update_stack(
            stack_identifier,
            template=self.template_param_has_immutable_field,
            parameters=update_parameters)
        stack = self.client.stacks.get(stack_identifier)

        # Verify the value of the updated parameter
        self.assertEqual(param1_update_value,
                         self._stack_output(stack, 'param1_output'))

        # Ensure stack is not in a failed state
        self.assertEqual('UPDATE_COMPLETE', stack.stack_status)

    def test_immutable_param_field_error(self):
        param1_create_value = 'value1'
        create_parameters = {"param1": param1_create_value}

        # Toggle the immutable field to preclude updating
        immutable_true = self.template_param_has_immutable_field.replace(
            'immutable: false', 'immutable: true')

        stack_identifier = self.stack_create(
            template=immutable_true,
            parameters=create_parameters
        )
        stack = self.client.stacks.get(stack_identifier)

        param1_update_value = 'value2'
        update_parameters = {"param1": param1_update_value}

        # Verify the value of the parameter
        self.assertEqual(param1_create_value,
                         self._stack_output(stack, 'param1_output'))

        # Attempt to update the stack with a new parameter value
        try:
            self.update_stack(
                stack_identifier,
                template=immutable_true,
                parameters=update_parameters)
        except heat_exceptions.HTTPBadRequest as exc:
            exp = ('The following parameters are immutable and may not be '
                   'updated: param1')
            self.assertIn(exp, str(exc))

        stack = self.client.stacks.get(stack_identifier)

        # Ensure stack is not in a failed state
        self.assertEqual('CREATE_COMPLETE', stack.stack_status)

        # Ensure immutable parameter has not changed
        self.assertEqual(param1_create_value,
                         self._stack_output(stack, 'param1_output'))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine.resources.openstack.neutron import port
from heat.engine.resources.openstack.neutron import router
from heat.engine import support
from heat.engine import translation

LOG = logging.getLogger(__name__)


class FloatingIP(neutron.NeutronResource):
    """A resource for managing Neutron floating ips.

    Floating IP addresses can change their association between routers by
    action of the user. One of the most common use cases for floating IPs is
    to provide public IP addresses to a private cloud, where there are a
    limited number of IP addresses available. Another is for a public cloud
    user to have a "static" IP address that can be reassigned when an instance
    is upgraded or moved.
    """

    entity = 'floatingip'

    PROPERTIES = (
        FLOATING_NETWORK_ID, FLOATING_NETWORK, FLOATING_SUBNET,
        VALUE_SPECS, PORT_ID, FIXED_IP_ADDRESS, FLOATING_IP_ADDRESS,
        DNS_NAME, DNS_DOMAIN,
    ) = (
        'floating_network_id', 'floating_network', 'floating_subnet',
        'value_specs', 'port_id', 'fixed_ip_address', 'floating_ip_address',
        'dns_name', 'dns_domain',
    )

    ATTRIBUTES = (
        ROUTER_ID, TENANT_ID, FLOATING_NETWORK_ID_ATTR, FIXED_IP_ADDRESS_ATTR,
        FLOATING_IP_ADDRESS_ATTR, PORT_ID_ATTR,
    ) = (
        'router_id', 'tenant_id', 'floating_network_id', 'fixed_ip_address',
        'floating_ip_address', 'port_id',
    )

    properties_schema = {
        FLOATING_NETWORK_ID: properties.Schema(
            properties.Schema.STRING,
            support_status=support.SupportStatus(
                status=support.HIDDEN,
                version='5.0.0',
                message=_('Use property %s.') % FLOATING_NETWORK,
                previous_status=support.SupportStatus(
                    status=support.DEPRECATED,
                    version='2014.2'
                )
            ),
            constraints=[
                constraints.CustomConstraint('neutron.network')
            ],
        ),
        FLOATING_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('Network to allocate floating IP from.'),
            support_status=support.SupportStatus(version='2014.2'),
            required=True,
            constraints=[
                constraints.CustomConstraint('neutron.network')
            ],
        ),
        FLOATING_SUBNET: properties.Schema(
            properties.Schema.STRING,
            _('Subnet to allocate floating IP from.'),
            support_status=support.SupportStatus(version='9.0.0'),
            constraints=[
                constraints.CustomConstraint('neutron.subnet')
            ],
        ),
        VALUE_SPECS: properties.Schema(
            properties.Schema.MAP,
            _('Extra parameters to include in the "floatingip" object in the '
              'creation request. Parameters are often specific to installed '
              'hardware or extensions.'),
            default={}
        ),
        PORT_ID: properties.Schema(
            properties.Schema.STRING,
            _('ID of an existing port with at least one IP address to '
              'associate with this floating IP.'),
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('neutron.port')
            ]
        ),
        FIXED_IP_ADDRESS: properties.Schema(
            properties.Schema.STRING,
            _('IP address to use if the port has multiple addresses.'),
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('ip_addr')
            ]
        ),
        FLOATING_IP_ADDRESS: properties.Schema(
            properties.Schema.STRING,
            _('IP address of the floating IP. NOTE: The default policy '
              'setting in Neutron restricts usage of this property to '
              'administrative users only.'),
            constraints=[
                constraints.CustomConstraint('ip_addr')
            ],
            support_status=support.SupportStatus(version='5.0.0'),
        ),
        DNS_NAME: properties.Schema(
            properties.Schema.STRING,
            _('DNS name associated with floating ip.'),
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('rel_dns_name')
            ],
            support_status=support.SupportStatus(version='7.0.0'),
        ),
        DNS_DOMAIN: properties.Schema(
            properties.Schema.STRING,
            _('DNS domain associated with floating ip.'),
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('dns_domain')
            ],
            support_status=support.SupportStatus(version='7.0.0'),
        ),
    }

    attributes_schema = {
        ROUTER_ID: attributes.Schema(
            _('ID of the router used as gateway, set when associated with a '
              'port.'),
            type=attributes.Schema.STRING
        ),
        TENANT_ID: attributes.Schema(
            _('The tenant owning this floating IP.'),
            type=attributes.Schema.STRING
        ),
        FLOATING_NETWORK_ID_ATTR: attributes.Schema(
            _('ID of the network in which this IP is allocated.'),
            type=attributes.Schema.STRING
        ),
        FIXED_IP_ADDRESS_ATTR: attributes.Schema(
            _('IP address of the associated port, if specified.'),
            type=attributes.Schema.STRING,
            cache_mode=attributes.Schema.CACHE_NONE
        ),
        FLOATING_IP_ADDRESS_ATTR: attributes.Schema(
            _('The allocated address of this IP.'),
            type=attributes.Schema.STRING
        ),
        PORT_ID_ATTR: attributes.Schema(
            _('ID of the port associated with this IP.'),
            type=attributes.Schema.STRING,
            cache_mode=attributes.Schema.CACHE_NONE
        ),
    }

    def translation_rules(self, props):
        client_plugin = self.client_plugin()
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.REPLACE,
                [self.FLOATING_NETWORK],
                value_path=[self.FLOATING_NETWORK_ID]
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.FLOATING_NETWORK],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_NETWORK
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.FLOATING_SUBNET],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_SUBNET
            )
        ]

    def _add_router_interface_dependencies(self, deps, resource):
        def port_on_subnet(resource, subnet):
            if not resource.has_interface('OS::Neutron::Port'):
                return False

            try:
                fixed_ips = resource.properties.get(port.Port.FIXED_IPS)
            except (ValueError, TypeError):
                # Properties errors will be caught later in validation, where
                # we can report them in their proper context.
                return False
            if not fixed_ips:
                # During create we have only unresolved value for
                # functions, so can not use None value for building
                # correct dependencies. Depend on all RouterInterfaces
                # when the port has no fixed IP specified, since we
                # can't safely assume that any are in different
                # networks.
                if subnet is None:
                    return True

                try:
                    p_net = (resource.properties.get(port.Port.NETWORK) or
                             resource.properties.get(port.Port.NETWORK_ID))
                except (ValueError, TypeError):
                    # Properties errors will be caught later in validation,
                    # where we can report them in their proper context.
                    return False
                if p_net:
                    try:
                        network = self.client().show_network(p_net)['network']
                        return subnet in network['subnets']
                    except Exception as exc:
                        LOG.info("Ignoring Neutron error while "
                                 "getting FloatingIP dependencies: %s",
                                 str(exc))
                        return False
            else:
                try:
                    fixed_ips = resource.properties.get(port.Port.FIXED_IPS)
                except (ValueError, TypeError):
                    # Properties errors will be caught later in validation,
                    # where we can report them in their proper context.
                    return False
                for fixed_ip in fixed_ips:
                    port_subnet = (fixed_ip.get(port.Port.FIXED_IP_SUBNET) or
                                   fixed_ip.get(port.Port.FIXED_IP_SUBNET_ID))
                    if subnet == port_subnet:
                        return True
            return False

        interface_subnet = (
            resource.properties.get(router.RouterInterface.SUBNET) or
            resource.properties.get(router.RouterInterface.SUBNET_ID))
        for d in deps.graph()[self]:
            if port_on_subnet(d, interface_subnet):
                deps += (self, resource)
                break

    def add_dependencies(self, deps):
        super(FloatingIP, self).add_dependencies(deps)

        for resource in self.stack.values():
            # depend on any RouterGateway in this template with the same
            # network_id as this floating_network_id
            if resource.has_interface('OS::Neutron::RouterGateway'):
                try:
                    gateway_network = (
                        resource.properties.get(router.RouterGateway.NETWORK)
                        or resource.properties.get(
                            router.RouterGateway.NETWORK_ID))
                    floating_network = self.properties[self.FLOATING_NETWORK]
                except (ValueError, TypeError):
                    # Properties errors will be caught later in validation,
                    # where we can report them in their proper context.
                    continue
                if gateway_network == floating_network:
                    deps += (self, resource)

            # depend on any RouterInterface in this template which interfaces
            # with the same subnet that this floating IP's port is assigned
            # to
            elif resource.has_interface('OS::Neutron::RouterInterface'):
                self._add_router_interface_dependencies(deps, resource)
            # depend on Router with EXTERNAL_GATEWAY_NETWORK property
            # this template with the same network_id as this
            # floating_network_id
            elif resource.has_interface('OS::Neutron::Router'):
                try:
                    gateway = resource.properties.get(
                        router.Router.EXTERNAL_GATEWAY)
                    floating_network = self.properties[self.FLOATING_NETWORK]
                except (ValueError, TypeError):
                    # Properties errors will be caught later in validation,
                    # where we can report them in their proper context.
                    continue
                if gateway:
                    gateway_network = gateway.get(
                        router.Router.EXTERNAL_GATEWAY_NETWORK)
                    if gateway_network == floating_network:
                        deps += (self, resource)

    def validate(self):
        super(FloatingIP, self).validate()
        # fixed_ip_address cannot be specified without a port_id
        if self.properties[self.PORT_ID] is None and self.properties[
                self.FIXED_IP_ADDRESS] is not None:
            raise exception.ResourcePropertyDependency(
                prop1=self.FIXED_IP_ADDRESS, prop2=self.PORT_ID)

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())
        props['floating_network_id'] = props.pop(self.FLOATING_NETWORK)
        if self.FLOATING_SUBNET in props:
            props['subnet_id'] = props.pop(self.FLOATING_SUBNET)
        fip = self.client().create_floatingip({
            'floatingip': props})['floatingip']
        self.resource_id_set(fip['id'])

    def handle_delete(self):
        with self.client_plugin().ignore_not_found:
            self.client().delete_floatingip(self.resource_id)
            return True

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            port_id = prop_diff.get(self.PORT_ID,
                                    self.properties[self.PORT_ID])

            fixed_ip_address = prop_diff.get(
                self.FIXED_IP_ADDRESS,
                self.properties[self.FIXED_IP_ADDRESS])

            request_body = {
                'floatingip': {
                    'port_id': port_id,
                    'fixed_ip_address': fixed_ip_address}}

            self.client().update_floatingip(self.resource_id, request_body)


class FloatingIPAssociation(neutron.NeutronResource):
    """A resource for associating floating ips and ports.

    This resource allows associating a floating IP to a port with at least one
    IP address to associate with this floating IP.
    """
    PROPERTIES = (
        FLOATINGIP_ID, PORT_ID, FIXED_IP_ADDRESS,
    ) = (
        'floatingip_id', 'port_id', 'fixed_ip_address',
    )

    properties_schema = {
        FLOATINGIP_ID: properties.Schema(
            properties.Schema.STRING,
            _('ID of the floating IP to associate.'),
            required=True,
            update_allowed=True
        ),
        PORT_ID: properties.Schema(
            properties.Schema.STRING,
            _('ID of an existing port with at least one IP address to '
              'associate with this floating IP.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('neutron.port')
            ]
        ),
        FIXED_IP_ADDRESS: properties.Schema(
            properties.Schema.STRING,
            _('IP address to use if the port has multiple addresses.'),
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('ip_addr')
            ]
        ),
    }

    def add_dependencies(self, deps):
        super(FloatingIPAssociation, self).add_dependencies(deps)

        for resource in self.stack.values():
            if resource.has_interface('OS::Neutron::RouterInterface'):

                def port_on_subnet(resource, subnet):
                    if not resource.has_interface('OS::Neutron::Port'):
                        return False
                    fixed_ips = resource.properties.get(
                        port.Port.FIXED_IPS) or []
                    for fixed_ip in fixed_ips:
                        port_subnet = (
                            fixed_ip.get(port.Port.FIXED_IP_SUBNET)
                            or fixed_ip.get(port.Port.FIXED_IP_SUBNET_ID))
                        return subnet == port_subnet
                    return False

                interface_subnet = (
                    resource.properties.get(router.RouterInterface.SUBNET) or
                    resource.properties.get(router.RouterInterface.SUBNET_ID))
                for d in deps.graph()[self]:
                    if port_on_subnet(d, interface_subnet):
                        deps += (self, resource)
                        break

    def handle_create(self):
        props = self.prepare_properties(self.properties, self.name)
        floatingip_id = props.pop(self.FLOATINGIP_ID)
        self.client().update_floatingip(floatingip_id, {
            'floatingip': props})
        self.resource_id_set(self.id)

    def handle_delete(self):
        if not self.resource_id:
            return

        with self.client_plugin().ignore_not_found:
            self.client().update_floatingip(
                self.properties[self.FLOATINGIP_ID],
                {'floatingip': {'port_id': None}})

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            floatingip_id = self.properties[self.FLOATINGIP_ID]
            port_id = self.properties[self.PORT_ID]
            # if the floatingip_id is changed, disassociate the port which
            # associated with the old floatingip_id
            if self.FLOATINGIP_ID in prop_diff:
                with self.client_plugin().ignore_not_found:
                    self.client().update_floatingip(
                        floatingip_id,
                        {'floatingip': {'port_id': None}})

            # associate the floatingip with the new port
            floatingip_id = (prop_diff.get(self.FLOATINGIP_ID) or
                             floatingip_id)
            port_id = prop_diff.get(self.PORT_ID) or port_id

            fixed_ip_address = (prop_diff.get(self.FIXED_IP_ADDRESS) or
                                self.properties[self.FIXED_IP_ADDRESS])

            request_body = {
                'floatingip': {
                    'port_id': port_id,
                    'fixed_ip_address': fixed_ip_address}}

            self.client().update_floatingip(floatingip_id, request_body)
            self.resource_id_set(self.id)


def resource_mapping():
    return {
        'OS::Neutron::FloatingIP': FloatingIP,
        'OS::Neutron::FloatingIPAssociation': FloatingIPAssociation,
    }



#    Copyright (c) 2018 AT&T Corporation.
#
#    All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.clients.os.neutron import neutron_constraints as nc

CLIENT_NAME = 'neutron'


class TapServiceConstraint(nc.NeutronExtConstraint):
    resource_name = 'tap_service'
    extension = 'taas'


class TapFlowConstraint(nc.NeutronExtConstraint):
    resource_name = 'tap_flow'
    extension = 'taas'


class TaaSProviderConstraint(nc.ProviderConstraint):
    service_type = 'TAPASASERVICE'



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""Service object."""

from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.common import service_utils
from heat.db.sqlalchemy import api as db_api
from heat.objects import base as heat_base


class Service(
        heat_base.HeatObject,
        base.VersionedObjectDictCompat,
        base.ComparableVersionedObject,
):
    fields = {
        'id': fields.StringField(),
        'engine_id': fields.StringField(),
        'host': fields.StringField(),
        'hostname': fields.StringField(),
        'binary': fields.StringField(),
        'topic': fields.StringField(),
        'report_interval': fields.IntegerField(),
        'created_at': fields.DateTimeField(read_only=True),
        'updated_at': fields.DateTimeField(nullable=True),
        'deleted_at': fields.DateTimeField(nullable=True)
    }

    @staticmethod
    def _from_db_object(context, service, db_service):
        for field in service.fields:
            service[field] = db_service[field]
        service._context = context
        service.obj_reset_changes()
        return service

    @classmethod
    def _from_db_objects(cls, context, list_obj):
        return [cls._from_db_object(context, cls(context), obj)
                for obj in list_obj]

    @classmethod
    def get_by_id(cls, context, service_id):
        service_db = db_api.service_get(context, service_id)
        service = cls._from_db_object(context, cls(), service_db)
        return service

    @classmethod
    def create(cls, context, values):
        return cls._from_db_object(
            context,
            cls(),
            db_api.service_create(context, values))

    @classmethod
    def update_by_id(cls, context, service_id, values):
        return cls._from_db_object(
            context,
            cls(),
            db_api.service_update(context, service_id, values))

    @classmethod
    def delete(cls, context, service_id, soft_delete=True):
        db_api.service_delete(context, service_id, soft_delete)

    @classmethod
    def get_all(cls, context):
        return cls._from_db_objects(context,
                                    db_api.service_get_all(context))

    @classmethod
    def get_all_by_args(cls, context, host, binary, hostname):
        return cls._from_db_objects(
            context,
            db_api.service_get_all_by_args(context,
                                           host,
                                           binary,
                                           hostname))

    @classmethod
    def active_service_count(cls, context):
        """Return the number of services reportedly active."""
        return len([
            svc for svc in cls.get_all(context)
            if service_utils.format_service(svc)['status'] == 'up'])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import weakref

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from stevedore import enabled

from heat.common import exception
from heat.common.i18n import _
from heat.common import pluginutils

LOG = logging.getLogger(__name__)


_default_backend = "heat.engine.clients.OpenStackClients"

cloud_opts = [
    cfg.StrOpt('cloud_backend',
               default=_default_backend,
               help=_("Fully qualified class name to use as "
                      "a client backend."))
]
cfg.CONF.register_opts(cloud_opts)


class OpenStackClients(object):
    """Convenience class to create and cache client instances."""

    def __init__(self, context):
        self._context = weakref.ref(context)
        self._clients = {}
        self._client_plugins = {}

    @property
    def context(self):
        ctxt = self._context()
        assert ctxt is not None, "Need a reference to the context"
        return ctxt

    def client_plugin(self, name):
        global _mgr
        if name in self._client_plugins:
            return self._client_plugins[name]
        if _mgr and name in _mgr.names():
            client_plugin = _mgr[name].plugin(self.context)
            self._client_plugins[name] = client_plugin
            return client_plugin

    def client(self, name, version=None):
        client_plugin = self.client_plugin(name)
        if client_plugin:
            if version:
                return client_plugin.client(version=version)
            else:
                return client_plugin.client()

        if name in self._clients:
            return self._clients[name]
        # call the local method _<name>() if a real client plugin
        # doesn't exist
        method_name = '_%s' % name
        if callable(getattr(self, method_name, None)):
            client = getattr(self, method_name)()
            self._clients[name] = client
            return client
        LOG.warning('Requested client "%s" not found', name)


class ClientBackend(object):
    """Class for delaying choosing the backend client module.

    Delay choosing the backend client module until the client's class needs
    to be initialized.
    """
    def __new__(cls, context):
        if cfg.CONF.cloud_backend == _default_backend:
            return OpenStackClients(context)
        else:
            try:
                return importutils.import_object(cfg.CONF.cloud_backend,
                                                 context)
            except (ImportError, RuntimeError, cfg.NoSuchOptError) as err:
                msg = _('Invalid cloud_backend setting in heat.conf '
                        'detected - %s') % str(err)
                LOG.error(msg)
                raise exception.Invalid(reason=msg)


Clients = ClientBackend


_mgr = None


def has_client(name):
    return _mgr and name in _mgr.names()


def initialise():
    global _mgr
    if _mgr:
        return

    def client_is_available(client_plugin):
        if not hasattr(client_plugin.plugin, 'is_available'):
            # if the client does not have a is_available() class method, then
            # we assume it wants to be always available
            return True
        # let the client plugin decide if it wants to register or not
        return client_plugin.plugin.is_available()

    _mgr = enabled.EnabledExtensionManager(
        namespace='heat.clients',
        check_func=client_is_available,
        invoke_on_load=False,
        on_load_failure_callback=pluginutils.log_fail_msg)


def list_opts():
    yield None, cloud_opts



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from heat.common import config
from heat.common import crypt
from heat.common import exception
from heat.tests import common


class CryptTest(common.HeatTestCase):

    def test_fernet_key(self):
        key = 'x' * 16
        method, result = crypt.encrypt('foo', key)
        self.assertEqual('cryptography_decrypt_v1', method)
        self.assertIsNotNone(result)

    def test_init_auth_encryption_key_length(self):
        """Test for length of the auth_encryption_length in config file"""
        cfg.CONF.set_override('auth_encryption_key', 'abcdefghijklma')
        err = self.assertRaises(exception.Error,
                                config.startup_sanity_check)
        exp_msg = ('heat.conf misconfigured, auth_encryption_key '
                   'must be 32 characters')
        self.assertIn(exp_msg, str(err))

    def _test_encrypt_decrypt_dict(self, encryption_key=None):
        data = {'p1': u'happy',
                '2': [u'a', u'little', u'blue'],
                'p3': {u'really': u'exited', u'ok int': 9},
                '4': u'',
                'p5': True,
                '6': 7}
        encrypted_data = crypt.encrypted_dict(data, encryption_key)
        for k in encrypted_data:
            self.assertEqual('cryptography_decrypt_v1',
                             encrypted_data[k][0])
            self.assertEqual(2, len(encrypted_data[k]))
        # the keys remain the same
        self.assertEqual(set(data), set(encrypted_data))

        decrypted_data = crypt.decrypted_dict(encrypted_data, encryption_key)
        self.assertEqual(data, decrypted_data)

    def test_encrypt_decrypt_dict_custom_enc_key(self):
        self._test_encrypt_decrypt_dict('just for testing not so great re')

    def test_encrypt_decrypt_dict_default_enc_key(self):
        self._test_encrypt_decrypt_dict()

    def test_decrypt_dict_invalid_key(self):
        data = {'p1': u'happy',
                '2': [u'a', u'little', u'blue'],
                '6': 7}
        encrypted_data = crypt.encrypted_dict(
            data, '767c3ed056cbaa3b9dfedb8c6f825bf0')
        ex = self.assertRaises(exception.InvalidEncryptionKey,
                               crypt.decrypted_dict,
                               encrypted_data,
                               '767c3ed056cbaa3b9dfedb8c6f825bf1')
        self.assertEqual('Can not decrypt data with the auth_encryption_key '
                         'in heat config.',
                         str(ex))



def check_resource_count(expected_count):
    test.assertEqual(expected_count, len(reality.all_resources()))

example_template = Template({
    'A': RsrcDef({}, []),
    'B': RsrcDef({'a': '4alpha'}, ['A']),
    'C': RsrcDef({'a': 'foo'}, ['B']),
    'D': RsrcDef({'a': 'bar'}, ['C']),
})
engine.create_stack('foo', example_template)
engine.noop(1)

example_template2 = Template({
    'A': RsrcDef({}, []),
    'B': RsrcDef({'a': '4alpha'}, ['A']),
    'C': RsrcDef({'a': 'blarg'}, ['B']),
    'D': RsrcDef({'a': 'wibble'}, ['C']),
})
engine.update_stack('foo', example_template2)
engine.call(check_resource_count, 2)
engine.noop(11)
engine.call(verify, example_template2)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support

from oslo_utils import timeutils


class Delay(resource.Resource):
    """A resource that pauses for a configurable delay.

    By manipulating the dependency relationships between resources in the
    template, a delay can be inserted at an arbitrary point during e.g. stack
    creation or deletion. They delay will occur after any resource that it
    depends on during CREATE or SUSPEND, and before any resource that it
    depends on during DELETE or RESUME. Similarly, it will occur before any
    resource that depends on it during CREATE or SUSPEND, and after any
    resource thet depends on it during DELETE or RESUME.

    If a non-zero maximum jitter is specified, a random amount of jitter -
    chosen with uniform probability in the range from 0 to the product of the
    maximum jitter value and the jitter multiplier (1s by default) - is added
    to the minimum delay time. This can be used, for example, in the scaled
    unit of a large scaling group to prevent 'thundering herd' issues.
    """

    support_status = support.SupportStatus(version='11.0.0')

    _ALLOWED_ACTIONS = (
        resource.Resource.CREATE,
        resource.Resource.DELETE,
        resource.Resource.SUSPEND,
        resource.Resource.RESUME,
    )

    PROPERTIES = (
        MIN_WAIT_SECS, MAX_JITTER, JITTER_MULTIPLIER_SECS, DELAY_ACTIONS,
    ) = (
        'min_wait', 'max_jitter', 'jitter_multiplier', 'actions',
    )

    properties_schema = {
        MIN_WAIT_SECS: properties.Schema(
            properties.Schema.NUMBER,
            _('Minimum time in seconds to wait during the specified actions.'),
            update_allowed=True,
            default=0,
            constraints=[
                constraints.Range(min=0)
            ]
        ),
        MAX_JITTER: properties.Schema(
            properties.Schema.NUMBER,
            _('Maximum jitter to add to the minimum wait time.'),
            update_allowed=True,
            default=0,
            constraints=[
                constraints.Range(min=0),
            ]
        ),
        JITTER_MULTIPLIER_SECS: properties.Schema(
            properties.Schema.NUMBER,
            _('Number of seconds to multiply the maximum jitter value by.'),
            update_allowed=True,
            default=1.0,
            constraints=[
                constraints.Range(min=0),
            ]
        ),
        DELAY_ACTIONS: properties.Schema(
            properties.Schema.LIST,
            _('Actions during which the delay will occur.'),
            update_allowed=True,
            default=[resource.Resource.CREATE],
            constraints=[constraints.AllowedValues(_ALLOWED_ACTIONS)]
        ),
    }

    attributes_schema = {}

    def _delay_parameters(self):
        """Return a tuple of the min delay and max jitter, in seconds."""
        min_wait_secs = self.properties[self.MIN_WAIT_SECS]
        max_jitter_secs = (self.properties[self.MAX_JITTER] *
                           self.properties[self.JITTER_MULTIPLIER_SECS])
        return min_wait_secs, max_jitter_secs

    def validate(self):
        result = super(Delay, self).validate()
        if not self.stack.strict_validate:
            return result

        min_wait_secs, max_jitter_secs = self._delay_parameters()
        max_wait = min_wait_secs + max_jitter_secs
        if max_wait > self.stack.timeout_secs():
            raise exception.StackValidationFailed(_('%(res_type)s maximum '
                                                    'delay %(max_wait)ss '
                                                    'exceeds stack timeout.') %
                                                  {'res_type': self.type,
                                                   'max_wait': max_wait})
        return result

    def _wait_secs(self, action):
        """Return a (randomised) wait time for the specified action."""
        if action not in self.properties[self.DELAY_ACTIONS]:
            return 0

        min_wait_secs, max_jitter_secs = self._delay_parameters()
        return min_wait_secs + (max_jitter_secs * random.random())

    def _handle_action(self):
        """Return a tuple of the start time in UTC and the time to wait."""
        return timeutils.utcnow(), self._wait_secs(self.action)

    @staticmethod
    def _check_complete(started_at, wait_secs):
        if not wait_secs:
            return True
        elapsed_secs = (timeutils.utcnow() - started_at).total_seconds()
        if elapsed_secs >= wait_secs:
            return True
        remaining = wait_secs - elapsed_secs
        if remaining >= 4:
            raise resource.PollDelay(int(remaining // 2))
        return False

    def handle_create(self):
        return self._handle_action()

    def handle_delete(self):
        return self._handle_action()

    def handle_suspend(self):
        return self._handle_action()

    def handle_resume(self):
        return self._handle_action()

    def check_create_complete(self, cookie):
        return self._check_complete(*cookie)

    def check_delete_complete(self, cookie):
        return self._check_complete(*cookie)

    def check_suspend_complete(self, cookie):
        return self._check_complete(*cookie)

    def check_resume_complete(self, cookie):
        return self._check_status_complete(*cookie)


def resource_mapping():
    return {
        'OS::Heat::Delay': Delay,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.engine.clients.os.keystone import fake_keystoneclient as fake_ks
from heat.engine import resource
from heat.engine.resources.openstack.keystone import domain
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

KEYSTONE_REGION_TEMPLATE = {
    'heat_template_version': '2017-02-24',
    'resources': {
        'test_domain': {
            'type': 'OS::Keystone::Domain',
            'properties': {
                'name': 'test_domain_1',
                'description': 'Test domain',
                'enabled': 'True'
            }
        }
    }
}


class KeystoneDomainTest(common.HeatTestCase):
    def setUp(self):
        super(KeystoneDomainTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack_keystone',
            template.Template(KEYSTONE_REGION_TEMPLATE)
        )

        self.test_domain = self.stack['test_domain']

        # Mock client
        self.keystoneclient = mock.Mock()
        self.patchobject(resource.Resource, 'client',
                         return_value=fake_ks.FakeKeystoneClient(
                             client=self.keystoneclient))
        self.domains = self.keystoneclient.domains

        keystone_client_plugin = mock.MagicMock()
        self.test_domain.client_plugin = mock.MagicMock()
        self.test_domain.client_plugin.return_value = keystone_client_plugin

    def _get_mock_domain(self):
        value = mock.MagicMock()
        domain_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        value.id = domain_id

        return value

    def test_domain_handle_create(self):
        mock_domain = self._get_mock_domain()
        self.domains.create.return_value = mock_domain

        # validate the properties
        self.assertEqual(
            'test_domain_1',
            self.test_domain.properties.get(domain.KeystoneDomain.NAME))
        self.assertEqual(
            'Test domain',
            self.test_domain.properties.get(
                domain.KeystoneDomain.DESCRIPTION))
        self.assertEqual(
            True,
            self.test_domain.properties.get(domain.KeystoneDomain.ENABLED))

        self.test_domain.handle_create()

        # validate domain creation
        self.domains.create.assert_called_once_with(
            name='test_domain_1',
            description='Test domain',
            enabled=True)

        # validate physical resource id
        self.assertEqual(mock_domain.id, self.test_domain.resource_id)

    def test_domain_handle_update(self):
        self.test_domain.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {domain.KeystoneDomain.DESCRIPTION:
                     'Test Domain updated',
                     domain.KeystoneDomain.ENABLED: False,
                     domain.KeystoneDomain.NAME: 'test_domain_2'}

        self.test_domain.handle_update(json_snippet=None,
                                       tmpl_diff=None,
                                       prop_diff=prop_diff)

        self.domains.update.assert_called_once_with(
            domain=self.test_domain.resource_id,
            description=prop_diff[domain.KeystoneDomain.DESCRIPTION],
            enabled=prop_diff[domain.KeystoneDomain.ENABLED],
            name='test_domain_2'
        )

    def test_get_live_state(self):
        sample_domain = {
            domain.KeystoneDomain.NAME: 'test',
            domain.KeystoneDomain.ENABLED: True,
            domain.KeystoneDomain.DESCRIPTION: 'test domain'
        }
        d = mock.Mock()
        d.to_dict.return_value = sample_domain
        self.domains.get.return_value = d

        reality = self.test_domain.get_live_state(self.test_domain.properties)
        self.assertEqual(sample_domain, reality)
